# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
"""

import warnings
warnings.warn(
    "All of datalad.cmdline is deprecated/discontinued as of datalad 0.16. "
    "A new CLI implementation is available at datalad.cli. "
    "Please adjust any imports.",
    DeprecationWarning)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""provide information about this DataLad installation"""

__docformat__ = 'restructuredtext'

import logging
import os
import os.path as op
import sys
import tempfile
from functools import partial


from datalad.interface.base import (
    Interface,
    build_doc,
)
from datalad.utils import (
    ensure_unicode,
    getpwd,
    unlink,
    Path,
)
from datalad.support.external_versions import external_versions
from datalad.support.exceptions import (
    CapturedException,
    CommandError,
    InvalidGitRepositoryError,
)
from datalad import __version__

lgr = logging.getLogger('datalad.local.wtf')


# wording to use for items which were considered sensitive and thus not shown
_HIDDEN = "<SENSITIVE, report disabled by configuration>"


# formatting helper
def _t2s(t):
    res = []
    for e in t:
        if isinstance(e, tuple):
            es = _t2s(e)
            if es != '':
                res += ['(%s)' % es]
        elif e != '':
            res += [e]
    return '/'.join(res)


def get_max_path_length(top_path=None, maxl=1000):
    """Deduce the maximal length of the filename in a given path
    """
    if not top_path:
        top_path = getpwd()
    import random
    from datalad import lgr
    from datalad.support import path
    prefix = path.join(top_path, "dl%d" % random.randint(1 ,100000))
    # some smart folks could implement binary search for this
    max_path_length = None
    for i in range(maxl-len(prefix)):
        filename = prefix + '_' * i
        path_length = len(filename)
        try:
            with open(filename, 'w') as f:
                max_path_length = path_length
        except Exception as exc:
            ce = CapturedException(exc)
            lgr.debug(
                "Failed to create sample file for length %d. Last succeeded was %s. Exception: %s",
                path_length, max_path_length, ce)
            break
        unlink(filename)
    return max_path_length


def _describe_datalad():

    return {
        'version': ensure_unicode(__version__),
    }


def _describe_annex():
    from datalad.cmd import (
        GitWitlessRunner,
        StdOutErrCapture,
    )

    runner = GitWitlessRunner()
    try:
        out = runner.run(
            ['git', 'annex', 'version'], protocol=StdOutErrCapture)
    except CommandError as e:
        ce = CapturedException(e)
        return dict(
            version='not available',
            message=ce.format_short(),
        )
    info = {}
    for line in out['stdout'].split(os.linesep):
        key = line.split(':')[0]
        if not key:
            continue
        value = line[len(key) + 2:].strip()
        key = key.replace('git-annex ', '')
        if key.endswith('s'):
            value = value.split()
        info[key] = value
    return info


def _describe_system():
    import platform as pl
    from datalad import get_encoding_info
    from datalad.utils import get_linux_distribution
    try:
        dist = get_linux_distribution()
    except Exception as exc:
        ce = CapturedException(exc)
        lgr.warning("Failed to get distribution information: %s", ce)
        dist = tuple()

    return {
        'type': os.name,
        'name': pl.system(),
        'release': pl.release(),
        'version': pl.version(),
        'distribution': ' '.join([_t2s(dist),
                                  _t2s(pl.mac_ver()),
                                  _t2s(pl.win32_ver())]).rstrip(),
        'max_path_length': get_max_path_length(getpwd()),
        'encoding': get_encoding_info(),
        'filesystem': {l: _get_fs_type(l, p) for l, p in
                       [('CWD', Path.cwd()),
                        ('TMP', Path(tempfile.gettempdir())),
                        ('HOME', Path.home())]}
    }


def _get_fs_type(loc, path, _was_warned=[]):
    """Return file system info for given Paths. Provide pathlib path as input"""
    res = {'path': path}
    try:
        from psutil import disk_partitions
        parts = {Path(p.mountpoint): p  for p in disk_partitions()}
        match = None
        for mp in parts:
            # if the mountpoint is the test path or its parent
            # take it, whenever there is no match, or a longer match
            if (mp == path or mp in path.parents) and (
                    match is None or len(mp.parents) > len(match.parents)):
                match = mp
        match = parts[match]
        for sattr, tattr in (('fstype', 'type'),
                             ('maxpath', 'max_pathlength'),
                             ('opts', 'mount_opts')):
            if hasattr(match, sattr):
                res[tattr] = getattr(match, sattr)
    except Exception as exc:
        ce = CapturedException(exc)
        # if an exception occurs, leave out the fs type. The result renderer can
        # display a hint based on its lack in the report
        if not _was_warned:
            (lgr.debug if isinstance(exc, ImportError) else lgr.warning) (
                "Could not determine filesystem types due to %s", ce)
            # Rely on side-effect of [] as default arg
            _was_warned.append("warned")
    return res


def _describe_environment():
    from datalad import get_envvars_info
    return get_envvars_info()


def _describe_python():
    import platform
    return {
        'version': platform.python_version(),
        'implementation': platform.python_implementation(),
    }


def _describe_configuration(cfg, sensitive):
    if not cfg:
        return _HIDDEN

    # make it into a dict to be able to reassign
    cfg = dict(cfg.items())

    if sensitive != 'all':
        # filter out some of the entries which known to be highly sensitive
        for k in cfg.keys():
            if 'user' in k or 'token' in k or 'passwd' in k:
                cfg[k] = _HIDDEN

    return cfg


def _describe_extensions():
    infos = {}
    from datalad.support.entrypoints import iter_entrypoints
    from importlib import import_module

    for ename, emod, eload in iter_entrypoints('datalad.extensions'):
        info = {}
        infos[ename] = info
        try:
            ext = eload()
            info['load_error'] = None
            info['description'] = ext[0]
            info['module'] = emod
            mod = import_module(emod, package='datalad')
            info['version'] = getattr(mod, '__version__', None)
        except Exception as e:
            ce = CapturedException(e)
            info['load_error'] = ce.format_short()
            continue
        info['entrypoints'] = entry_points = {}
        for ep in ext[1]:
            ep_info = {
                'module': ep[0],
                'class': ep[1],
                'names': ep[2:],
            }
            entry_points['{}.{}'.format(*ep[:2])] = ep_info
            try:
                import_module(ep[0], package='datalad')
                ep_info['load_error'] = None
            except Exception as e:
                ce = CapturedException(e)
                ep_info['load_error'] = ce.format_short()
                continue
    return infos


def _describe_metadata_elements(group):
    infos = {}
    from datalad.support.entrypoints import iter_entrypoints
    from importlib import import_module
    if sys.version_info < (3, 10):
        # 3.10 is when it was no longer provisional
        from importlib_metadata import distribution
    else:
        from importlib.metadata import distribution


    for ename, emod, eload in iter_entrypoints(group):
        info = {}
        infos[f'{ename}'] = info
        try:
            info['module'] = emod
            dist = distribution(emod.split('.', maxsplit=1)[0])
            info['distribution'] = f'{dist.name} {dist.version}'
            mod = import_module(emod, package='datalad')
            version = getattr(mod, '__version__', None)
            if version:
                # no not clutter the report with no version
                info['version'] = version
            eload()
            info['load_error'] = None
        except Exception as e:
            ce = CapturedException(e)
            info['load_error'] = ce.format_short()
            continue
    return infos


def _describe_dependencies():
    return {
        k: str(external_versions[k]) for k in external_versions.keys(query=True)
    }


def _describe_dataset(ds, sensitive):
    from datalad.interface.results import success_status_map

    try:
        infos = {
            'path': ds.path,
            'repo': ds.repo.__class__.__name__ if ds.repo else None,
            'id': ds.id,
        }
        # describe available branches and their states
        branches = [
            '%s@%s' % (b, next(ds.repo.get_branch_commits_(branch=b))[:7])
            for b in ds.repo.get_branches()]
        infos['branches'] = branches
        return infos
    except InvalidGitRepositoryError as e:
        ce = CapturedException(e)
        return {"invalid": ce.message}


def _describe_location(res):
    return {
        'path': res['path'],
        'type': res['type'],
    }


def _describe_credentials():
    import keyring
    from keyring.util import platform_

    def describe_keyring_backend(be):
        be_repr = repr(be)
        return be.name if 'object at 0' in be_repr else be_repr.strip('<>')

    # might later add information on non-keyring credentials gh-4981
    props = {}

    active_keyring = keyring.get_keyring()
    krp = {
        'config_file': Path(platform_.config_root(), 'keyringrc.cfg'),
        'data_root': platform_.data_root(),
        'active_backends': [
            describe_keyring_backend(be)
            for be in getattr(active_keyring, 'backends', [active_keyring])
        ],
    }
    props.update(
        keyring=krp,
    )
    return props


# Actual callables for WTF. If None -- should be bound later since depend on
# the context
SECTION_CALLABLES = {
    'datalad': _describe_datalad,
    'python': _describe_python,
    'git-annex': _describe_annex,
    'system': _describe_system,
    'environment': _describe_environment,
    'configuration': None,
    'location': None,
    'extensions': _describe_extensions,
    'dependencies': _describe_dependencies,
    'dataset': None,
    'credentials': _describe_credentials,
}


@build_doc
class WTF(Interface):
    """Generate a report about the DataLad installation and configuration

    IMPORTANT: Sharing this report with untrusted parties (e.g. on the web)
    should be done with care, as it may include identifying information, and/or
    credentials or access tokens.
    """
    result_renderer = 'tailored'

    from datalad.support.param import Parameter
    from datalad.distribution.dataset import datasetmethod
    from datalad.interface.base import eval_results
    from datalad.distribution.dataset import EnsureDataset
    from datalad.support.constraints import EnsureNone, EnsureChoice

    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc=""""specify the dataset to report on.
            no dataset is given, an attempt is made to identify the dataset
            based on the current working directory.""",
            constraints=EnsureDataset() | EnsureNone()),
        sensitive=Parameter(
            args=("-s", "--sensitive",),
            constraints=EnsureChoice(None, 'some', 'all'),
            doc="""if set to 'some' or 'all', it will display sections such as 
            config and metadata which could potentially contain sensitive 
            information (credentials, names, etc.).  If 'some', the fields
            which are known to be sensitive will still be masked out"""),
        sections=Parameter(
            args=("-S", "--section"),
            action='append',
            dest='sections',
            metavar="SECTION",
            constraints=EnsureChoice(None, *sorted(SECTION_CALLABLES) + ['*']),
            doc="""section to include.  If not set - depends on flavor.
            '*' could be used to force all sections.
            [CMD: This option can be given multiple times. CMD]"""),
        flavor=Parameter(
            args=("--flavor",),
            constraints=EnsureChoice('full', 'short'),
            doc="""Flavor of WTF. 'full' would produce markdown with exhaustive list of sections.
            'short' will provide a condensed summary only of datalad and dependencies by default.
            Use [CMD: --section CMD][PY: `section` PY] to list other sections"""),
        decor=Parameter(
            args=("-D", "--decor"),
            constraints=EnsureChoice('html_details', None),
            doc="""decoration around the rendering to facilitate embedding into
            issues etc, e.g. use 'html_details' for posting collapsible entry
            to GitHub issues."""),
        clipboard=Parameter(
            args=("-c", "--clipboard",),
            action="store_true",
            doc="""if set, do not print but copy to clipboard (requires pyperclip
            module)"""),
    )

    @staticmethod
    @datasetmethod(name='wtf')
    @eval_results
    def __call__(*, dataset=None, sensitive=None, sections=None, flavor="full", decor=None, clipboard=None):
        from datalad.distribution.dataset import require_dataset
        from datalad.support.exceptions import NoDatasetFound
        from datalad.interface.results import get_status_dict

        ds = None
        try:
            ds = require_dataset(dataset, check_installed=False, purpose='report')
        except NoDatasetFound:
            # failure is already logged
            pass
        if ds and not ds.is_installed():
            # warn that the dataset is bogus
            yield dict(
                action='wtf',
                path=ds.path,
                status='impossible',
                message=(
                    'No dataset found at %s. Reporting on the dataset is '
                    'not attempted.', ds.path),
                logger=lgr
            )
            # we don't deal with absent datasets
            ds = None
        if sensitive:
            if ds is None:
                from datalad import cfg
            else:
                cfg = ds.config
        else:
            cfg = None

        from datalad.ui import ui
        from datalad.support.external_versions import external_versions

        infos = dict()
        res = get_status_dict(
            action='wtf',
            path=ds.path if ds else ensure_unicode(op.abspath(op.curdir)),
            type='dataset' if ds else 'directory',
            status='ok',
            logger=lgr,
            decor=decor,
            infos=infos,
            flavor=flavor,
        )

        # Define section callables which require variables.
        # so there is no side-effect on module level original
        section_callables = SECTION_CALLABLES.copy()
        section_callables['location'] = partial(_describe_location, res)
        section_callables['configuration'] = \
            partial(_describe_configuration, cfg, sensitive)
        if ds:
            section_callables['dataset'] = \
                partial(_describe_dataset, ds, sensitive)
        else:
            section_callables.pop('dataset')
        assert all(section_callables.values())  # check if none was missed

        asked_for_all_sections = sections is not None and any(s == '*' for s in sections)
        if sections is None or asked_for_all_sections:
            if flavor == 'full' or asked_for_all_sections:
                sections = sorted(list(section_callables))
            elif flavor == 'short':
                sections = ['datalad', 'dependencies']
            else:
                raise ValueError(flavor)

        for s in sections:
            try:
                infos[s] = section_callables[s]()
            except KeyError:
                yield get_status_dict(
                    action='wtf',
                    path=Path.cwd(),
                    status='impossible',
                    message=(
                      'Requested section <%s> was not found among the '
                      'available infos. Skipping report.', s),
                    logger=lgr
                    )
        if clipboard:
            external_versions.check(
                'pyperclip', msg="It is needed to be able to use clipboard")
            import pyperclip
            report = _render_report(res)
            pyperclip.copy(report)
            ui.message("WTF information of length %s copied to clipboard"
                       % len(report))
        yield res
        return

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        from datalad.ui import ui
        out = _render_report(res)
        ui.message(out)
        # add any necessary hints afterwards
        maybe_show_hints(res)


def maybe_show_hints(res):
    """Helper to add hints to custom result rendering"""
    # check for missing file system info
    # if the system record lacks file system information, hint at a psutil
    # problem. Query for TMP should be safe, is included in system info
    from datalad.ui.utils import show_hint

    if 'system' in res.get('infos', {}) and \
        'type' not in res['infos']['system']['filesystem']['TMP']:
        try:
            import psutil
        except ImportError:
            show_hint('Hint: install psutil to get filesystem information')


def _render_report(res):
    report = u'# WTF'

    def _unwind(text, val, top):
        if isinstance(val, dict):
            for k in sorted(val):
                text += u'\n{}{} {}{} '.format(
                    '##' if not top else top,
                    '-' if top else '',
                    k,
                    ':' if top else '')
                text = _unwind(text, val[k], u'{}  '.format(top))
        elif isinstance(val, (list, tuple)):
            for i, v in enumerate(val):
                text += u'\n{}{} '.format(top, '-')
                text = _unwind(text, v, u'{}  '.format(top))
        else:
            text += u'{}'.format(val)
        return text

    def _unwind_short(text, val, top):
        if isinstance(val, dict):
            if not top:
                text += '\n'
                for k, v in val.items():
                    text += "- " + _unwind_short(k, v, top + ' ') + '\n'
            else:
                text += ": " + ' '.join('%s=%s' % i for i in val.items())
        elif isinstance(val, (list, tuple)):
            text += (' ' if not top else '\n').join(map(str, val))
        else:
            text += u'{}'.format(val)
        return text

    unwinder = {'full': _unwind, 'short': _unwind_short}[res.get('flavor', 'full')]
    report = unwinder(report, res.get('infos', {}), '')
    decor = res.get('decor', None)

    if not decor:
        return report

    if decor == 'html_details':
        report = """\
<details><summary>DataLad %s WTF (%s)</summary>

%s
</details>
        """ % (__version__, ', '.join(res.get('infos', {})), report)
    else:
        raise ValueError("Unknown value of decor=%s" % decor)
    return report



# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test foreach-dataset command"""

import os.path as op
import sys
from pathlib import Path

import pytest

from datalad.api import create
from datalad.distribution.dataset import Dataset
from datalad.tests.utils_pytest import (
    assert_false,
    assert_greater,
    assert_in,
    assert_not_in,
    assert_status,
    eq_,
    get_deeply_nested_structure,
    ok_clean_git,
    swallow_outputs,
    with_tempfile,
)


def _without_command(results):
    """A helper to tune up results so that they lack 'command'
    which is guaranteed to differ between different cmd types
    """
    out = []
    for r in results:
        r = r.copy()
        r.pop('command')
        out.append(r)
    return out


@with_tempfile(mkdir=True)
def check_basic_resilience(populator, path=None):
    ds = populator(path)
    ds.save()
    kwargs = dict(recursive=True)

    res_external = ds.foreach_dataset(
        [sys.executable, '-c', 'from datalad.distribution.dataset import Dataset; ds=Dataset("."); print(ds.path)'],
        **kwargs)
    res_python = ds.foreach_dataset("ds.path", cmd_type='eval', **kwargs)

    # a sample python function to pass to foreach
    def get_path(ds, **kwargs):
        return ds.path

    res_python_func = ds.foreach_dataset(get_path, **kwargs)

    assert_status('ok', res_external)
    assert_status('ok', res_python)

    # consistency checks
    eq_(len(res_external), len(res_python))
    eq_(len(res_external), len(res_python_func))
    eq_(_without_command(res_python), _without_command(res_python_func))

    # Test correct order for bottom-up vs top-down
    topdown_dss = [ds.path] + ds.subdatasets(result_xfm='paths', bottomup=False, **kwargs)
    eq_(topdown_dss, [_['result'] for _ in res_python])

    bottomup_dss = ds.subdatasets(result_xfm='paths', recursive=True, bottomup=True) + [ds.path]
    eq_(bottomup_dss, [_['result'] for _ in ds.foreach_dataset("ds.path", bottomup=True, cmd_type='eval', **kwargs)])

    # more radical example - cleanup
    # Make all datasets dirty
    for d in bottomup_dss:
        (Path(d) / "dirt").write_text("")
    res_clean = ds.foreach_dataset(['git', 'clean', '-f'], jobs=10, **kwargs)
    assert_status('ok', res_clean)
    # no dirt should be left
    for d in bottomup_dss:
        assert_false((Path(d) / "dirt").exists())

    if populator is get_deeply_nested_structure:
        ok_clean_git(ds.path, index_modified=[ds.pathobj / 'subds_modified'])
    else:
        ok_clean_git(ds.path)


@pytest.mark.parametrize("populator", [
    # empty dataset
    create,
    # ver much not empty dataset
    get_deeply_nested_structure,
])
def test_basic_resilience(populator):
    check_basic_resilience(populator)


@with_tempfile(mkdir=True)
def check_python_eval(cmd, path):
    ds = Dataset(path).create()
    res = ds.foreach_dataset(cmd, cmd_type='eval')
    eq_(len(res), 1)
    expected_variables = {'ds', 'pwd', 'refds'}
    eq_(expected_variables.intersection(res[0]['result']), expected_variables)
    # besides expected, there could be few more ATM, +5 arbitrarily just to test
    # that we are not leaking too much
    assert_greater(len(expected_variables) + 5, len(res[0]['result']))


@with_tempfile(mkdir=True)
def check_python_exec(cmd, path):
    ds = Dataset(path).create()
    sub = ds.create('sub')  # create subdataset for better coverage etc

    # but exec has no result
    res = ds.foreach_dataset(cmd, cmd_type='exec')
    assert_not_in('result', res[0])

    # but allows for more complete/interesting setups in which we could import modules etc
    cmd2 = 'import os, sys; print(f"DIR: {os.linesep.join(dir())}")'
    with swallow_outputs() as cmo:
        res1 = ds.foreach_dataset(cmd2, output_streams='capture', cmd_type='exec')
        assert_in('ds', res1[0]['stdout'])
        assert_in('sys', res1[0]['stdout'])
        eq_(res1[0]['stderr'], '')
        # default renderer for each dataset
        assert cmo.out.startswith(f'foreach-dataset(ok): {path}')
        assert f'foreach-dataset(ok): {sub.path}' in cmo.out

    with swallow_outputs() as cmo:
        res2 = ds.foreach_dataset(cmd2, output_streams='relpath', cmd_type='exec')
        # still have the same res
        assert res1 == res2
        # but we have "fancier" output
        assert cmo.out.startswith(f'DIR: ')
        # 2nd half should be identical to 1st half but with lines prefixed with sub/ path
        lines = cmo.out.splitlines()
        half = len(lines) // 2
        assert [op.join('sub', l) for l in lines[:half]] == lines[half:]
        assert 'foreach-dataset(ok)' not in cmo.out


def test_python():
    check_python_eval("dir()")
    check_python_exec("dir()")

    def dummy_dir(*args, **kwargs):
        """Ensure that we pass all placeholders as kwargs"""
        assert not args
        return kwargs

    check_python_eval(dummy_dir)  # direct function invocation



# ex: set sts=4 ts=4 sw=4 noet:
# -*- coding: utf-8 -*-
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""

"""

from os.path import join as opj

from packaging.version import Version

import datalad
from datalad.distribution.dataset import Dataset
from datalad.tests.utils_pytest import (
    assert_in,
    assert_in_results,
    assert_not_in,
    assert_raises,
    assert_result_count,
    swallow_outputs,
    with_tempfile,
    with_tree,
)

# before 0.14 ui.message() (used in dump) was not friendly to unicode
complicated_str = 'complicated {} beast with.dot'.format(
    "の" if Version(datalad.__version__) >= Version('0.14.0') else 'blob'
)

_config_file_content = """\
[something "user"]
name = Jane Doe
email = jd@example.com
novalue
empty =
myint = 3

[onemore "{}"]
findme = 5.0
""".format(complicated_str)

_dataset_config_template = {
    'ds': {
        '.datalad': {
            'config': _config_file_content}}}


@with_tree(tree=_dataset_config_template)
@with_tempfile(mkdir=True)
def test_something(path=None, new_home=None):
    ds = Dataset(opj(path, 'ds')).create(force=True)
    ds.save()

    # catches unsupported argument combinations
    assert_raises(ValueError, ds.configuration, 'dump', scope='branch')
    assert_raises(ValueError, ds.configuration, 'set', spec=('onlyname',))
    assert_raises(ValueError, ds.configuration, 'set', spec='nosection=value')
    # we also get that from the internal helper
    from datalad.local.configuration import configuration as cfghelper
    assert_in_results(
        cfghelper('set', 'global', [('nosection', 'value')], {}),
        status='error',
    )
    assert_raises(ValueError, ds.configuration, 'invalid')
    res = ds.configuration(result_renderer='disabled')

    assert_in_results(
        res,
        name='something.user.name',
        value='Jane Doe')
    # UTF handling
    assert_in_results(
        res,
        name=u'onemore.{}.findme'.format(complicated_str),
        value='5.0')

    res = ds.configuration(
        'set',
        spec='some.more=test',
        result_renderer='disabled',
    )
    assert_in_results(
        res,
        name='some.more',
        value='test')
    # Python tuple specs
    # swallow outputs to be able to execise the result renderer
    with swallow_outputs():
        res = ds.configuration(
            'set',
            spec=[
                ('some.more.still', 'test2'),
                # value is non-str -- will be converted
                ('lonely.val', 4)],
        )
    assert_in_results(
        res,
        name='some.more.still',
        value='test2')
    assert_in_results(
        res,
        name='lonely.val',
        value='4')

    assert_in_results(
        ds.configuration('get', spec='lonely.val'),
        status='ok',
        name='lonely.val',
        value='4',
    )

    # remove something that does not exist in the specified scope
    assert_in_results(
        ds.configuration('unset', scope='branch', spec='lonely.val',
                         result_renderer='disabled', on_failure='ignore'),
        status='error')
    # remove something that does not exist in the specified scope
    assert_in_results(
        ds.configuration('unset', spec='lonely.val',
                         result_renderer='disabled'),
        status='ok')
    assert_not_in('lonely.val', ds.config)
    # errors if done again
    assert_in_results(
        ds.configuration('unset', spec='lonely.val',
                         result_renderer='disabled', on_failure='ignore'),
        status='error')

    # add a subdataset to test recursive operation
    subds = ds.create('subds')

    with swallow_outputs():
        res = ds.configuration('set', spec='rec.test=done', recursive=True)
    assert_result_count(
        res,
        2,
        name='rec.test',
        value='done',
    )

    # exercise the result renderer
    with swallow_outputs() as cml:
        ds.configuration(recursive=True)
        # we get something on the subds with the desired markup
        assert_in('<ds>/subds:rec.test=done', cml.out)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Distribution utility functions

"""

import logging

from os.path import (
    isabs,
    join as opj,
    normpath,
)
import posixpath
from datalad.log import log_progress
from datalad.support.annexrepo import AnnexRepo
from datalad.support.network import (
    PathRI,
    RI,
    URL,
)


lgr = logging.getLogger('datalad.distribution.utils')


def _get_flexible_source_candidates(src, base_url=None, alternate_suffix=True):
    """Get candidates to try cloning from.

    Primarily to mitigate the problem that git doesn't append /.git
    while cloning from non-bare repos over dummy protocol (http*).  Also to
    simplify creation of urls whenever base url and relative path within it
    provided

    Parameters
    ----------
    src : string or RI
      Full or relative (then considered within base_url if provided) path
    base_url : string or RI, optional
    alternate_suffix : bool
      Whether to generate URL candidates with and without '/.git' suffixes.

    Returns
    -------
    candidates : list of str
      List of RIs (path, url, ssh targets) to try to install from
    """
    candidates = []

    ri = RI(src)
    if isinstance(ri, PathRI) and not isabs(ri.path) and base_url:
        ri = RI(base_url)
        if ri.path.endswith('/.git'):
            base_path = ri.path[:-5]
            base_suffix = '.git'
        else:
            base_path = ri.path
            base_suffix = ''
        if isinstance(ri, PathRI):
            # this is a path, so stay native
            ri.path = normpath(opj(base_path, src, base_suffix))
        else:
            # we are handling a URL, use POSIX path conventions
            ri.path = posixpath.normpath(
                posixpath.join(base_path, src, base_suffix))

    src = str(ri)

    candidates.append(src)
    if alternate_suffix and isinstance(ri, URL):
        if ri.scheme in {'http', 'https'}:
            # additionally try to consider .git:
            if not src.rstrip('/').endswith('/.git'):
                candidates.append(
                    '{0}/.git'.format(src.rstrip('/')))

    return candidates



def _yield_ds_w_matching_siblings(
        ds, names, recursive=False, recursion_limit=None):
    """(Recursively) inspect a dataset for siblings with particular name(s)

    Parameters
    ----------
    ds: Dataset
      The dataset to be inspected.
    names: iterable
      Sibling names (str) to test for.
    recursive: bool, optional
      Whether to recurse into subdatasets.
    recursion_limit: int, optional
      Recursion depth limit.

    Yields
    ------
    str, str
      Path to the dataset with a matching sibling, and name of the matching
      sibling in that dataset.
    """

    def _discover_all_remotes(ds, refds, **kwargs):
        """Helper to be run on all relevant datasets via foreach
        """
        # Note, that `siblings` doesn't tell us about not enabled special
        # remotes. There could still be conflicting names we need to know
        # about in order to properly deal with the `existing` switch.

        repo = ds.repo
        # list of known git remotes
        if isinstance(repo, AnnexRepo):
            remotes = repo.get_remotes(exclude_special_remotes=True)
            remotes.extend([v['name']
                            for k, v in repo.get_special_remotes().items()]
                           )
        else:
            remotes = repo.get_remotes()
        return remotes

    if not recursive:
        for name in _discover_all_remotes(ds, ds):
            if name in names:
                yield ds.path, name
        return

    # in recursive mode this check could take a substantial amount of
    # time: employ a progress bar (or rather a counter, because we don't
    # know the total in advance
    pbar_id = 'check-siblings-{}'.format(id(ds))
    log_progress(
        lgr.info, pbar_id,
        'Start checking pre-existing sibling configuration %s', ds,
        label='Query siblings',
        unit=' Siblings',
    )

    for res in ds.foreach_dataset(
            _discover_all_remotes,
            recursive=recursive,
            recursion_limit=recursion_limit,
            return_type='generator',
            result_renderer='disabled',
    ):
        # unwind result generator
        if 'result' in res:
            for name in res['result']:
                log_progress(
                    lgr.info, pbar_id,
                    'Discovered sibling %s in dataset at %s',
                    name, res['path'],
                    update=1,
                    increment=True)
                if name in names:
                    yield res['path'], name

    log_progress(
        lgr.info, pbar_id,
        'Finished checking pre-existing sibling configuration %s', ds,
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-; coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""test command datalad {re,}run

"""

__docformat__ = 'restructuredtext'

import logging
import os.path as op
import sys
from io import StringIO
from os import remove

import pytest
from unittest.mock import patch

from datalad.api import (
    install,
    run,
)
from datalad.core.local.run import run_command
from datalad.core.local.tests.test_run import last_commit_msg
from datalad.distribution.dataset import Dataset
from datalad.local.rerun import (
    diff_revision,
    get_run_info,
    new_or_modified,
)
from datalad.support.exceptions import (
    CommandError,
    IncompleteResultsError,
)
from datalad.support.gitrepo import GitRepo
from datalad.tests.utils_pytest import (
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    SkipTest,
    assert_dict_equal,
    assert_false,
    assert_in,
    assert_in_results,
    assert_not_in,
    assert_not_in_results,
    assert_raises,
    assert_repo_status,
    assert_result_count,
    assert_status,
    create_tree,
    eq_,
    known_failure_windows,
    neq_,
    ok_,
    ok_exists,
    ok_file_has_content,
    ok_file_under_git,
    skip_if_adjusted_branch,
    slow,
    swallow_logs,
    swallow_outputs,
    with_tempfile,
    with_tree,
)
from datalad.utils import (
    Path,
    chpwd,
    on_windows,
)

cat_command = 'cat' if not on_windows else 'type'
touch_command = "touch " if not on_windows else "type nul > "
grep_command = 'grep ' if not on_windows else 'findstr '

@slow  # 17.1880s
@known_failure_windows
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_rerun(path=None, nodspath=None):
    ds = Dataset(path).create()
    sub = ds.create('sub')
    probe_path = op.join(sub.path, 'sequence')
    # run inside the dataset
    with chpwd(path), \
            swallow_outputs():
        ds.run(f'echo x$({cat_command} sub/sequence) > sub/sequence')
    # command ran once, all clean
    assert_repo_status(ds.path)
    eq_('x\n', open(probe_path).read())
    # now, for a rerun we can be anywhere, PWD and all are recorded
    # moreover, rerun must figure out which bits to unlock, even in
    # subdatasets
    with chpwd(nodspath), \
            swallow_outputs():
        ds.rerun()
    assert_repo_status(ds.path)
    # ran twice now
    eq_('xx\n', open(probe_path).read())

    # Rerunning from a subdataset skips the command.
    _, sub_info = get_run_info(ds, last_commit_msg(sub.repo))
    eq_(ds.id, sub_info["dsid"])
    assert_result_count(
        sub.rerun(return_type="list", on_failure="ignore"),
        1, status="impossible", action="run", rerun_action="skip")
    eq_('xx\n', open(probe_path).read())

    # Rerun fails with a dirty repo.
    dirt = op.join(path, "dirt")
    with open(dirt, "w") as fh:
        fh.write("")
    assert_status('impossible', ds.rerun(on_failure="ignore"))
    remove(dirt)
    assert_repo_status(ds.path)

    # Make a non-run commit.
    with open(op.join(path, "nonrun-file"), "w") as f:
        f.write("foo")
    ds.save("nonrun-file")
    # Now rerun the buried command.
    ds.rerun(revision=DEFAULT_BRANCH + "~", message="rerun buried")
    eq_('xxx\n', open(probe_path).read())
    # Also check that the messasge override worked.
    eq_(last_commit_msg(ds.repo).splitlines()[0],
        "[DATALAD RUNCMD] rerun buried")
    # Or a range of commits, skipping non-run commits.
    ds.rerun(since=DEFAULT_BRANCH + "~3")
    eq_('xxxxx\n', open(probe_path).read())
    # Or --since= to run all reachable commits.
    ds.rerun(since="")
    eq_('xxxxxxxxxx\n', open(probe_path).read())

    # We can get back a report of what would happen rather than actually
    # rerunning anything.
    report = ds.rerun(since="", report=True, return_type="list")
    # The "diff" section of the report doesn't include the unchanged files that
    # would come in "-f json diff" output.
    for entry in report:
        if entry["rerun_action"] == "run":
            # None of the run commits touch .datalad/config or any other config
            # file.
            assert_false(any(r["path"].endswith("config")
                             for r in entry["diff"]))

    # Nothing changed.
    eq_('xxxxxxxxxx\n', open(probe_path).read())
    assert_result_count(report, 1, rerun_action="skip-or-pick")
    report[-1]["commit"] == ds.repo.get_hexsha()

    # If a file is dropped, we remove it instead of unlocking it.
    ds.drop(probe_path, reckless='kill')
    with swallow_outputs():
        ds.rerun()

    eq_('x\n', open(probe_path).read())


@with_tempfile(mkdir=True)
def test_rerun_empty_branch(path=None):
    GitRepo(path, create=True)
    ds = Dataset(path)
    assert_status("impossible", ds.rerun(on_failure="ignore"))


@with_tempfile(mkdir=True)
def test_rerun_onto(path=None):
    ds = Dataset(path).create()
    if ds.repo.is_managed_branch():
        assert_status('impossible',
                      ds.rerun(onto="triggers-abort", on_failure="ignore"))
        raise SkipTest("Test incompatible with adjusted branch")

    # Make sure we have more than one commit. The one commit case is checked
    # elsewhere.
    ds.repo.commit(msg="noop commit", options=["--allow-empty"])

    grow_file = op.join(path, "grows")

    # Make sure we can handle range-specifications that yield no results.
    for since in ["", "HEAD"]:
        assert_result_count(
            ds.rerun("HEAD", onto="", since=since, on_failure="ignore"),
            1, status="impossible", action="run")

    ds.run('echo static-content > static')
    ds.repo.tag("static")
    with swallow_outputs():
        ds.run(f'echo x$({cat_command} grows) > grows')
    ds.rerun()
    eq_('xx\n', open(grow_file).read())

    # If we run the "static" change on top of itself, we end up in the
    # same (but detached) place.
    ds.rerun(revision="static", onto="static")
    ok_(ds.repo.get_active_branch() is None)
    eq_(ds.repo.get_hexsha(),
        ds.repo.get_hexsha("static"))

    # If we run the "static" change from the same "base", we end up
    # with a new commit.
    ds.repo.checkout(DEFAULT_BRANCH)
    with swallow_outputs():
        ds.rerun(revision="static", onto="static^")
    ok_(ds.repo.get_active_branch() is None)
    neq_(ds.repo.get_hexsha(),
         ds.repo.get_hexsha("static"))
    ok_(all(r["state"] == "clean" for r in ds.diff(fr="HEAD", to="static")))
    for revrange in ["..static", "static.."]:
        eq_(len(ds.repo.get_revisions(revrange)), 1)

    # Unlike the static change, if we run the ever-growing change on
    # top of itself, we end up with a new commit.
    ds.repo.checkout(DEFAULT_BRANCH)
    ds.rerun(onto="HEAD")
    ok_(ds.repo.get_active_branch() is None)
    neq_(ds.repo.get_hexsha(),
         ds.repo.get_hexsha(DEFAULT_BRANCH))

    # An empty `onto` means use the parent of the first revision.
    ds.repo.checkout(DEFAULT_BRANCH)
    with swallow_outputs():
        ds.rerun(since="static^", onto="")
    ok_(ds.repo.get_active_branch() is None)
    for revrange in [".." + DEFAULT_BRANCH, DEFAULT_BRANCH + ".."]:
        eq_(len(ds.repo.get_revisions(revrange)), 3)

    # An empty `onto` means use the parent of the first revision that
    # has a run command.
    ds.repo.checkout(DEFAULT_BRANCH)
    with swallow_outputs():
        ds.rerun(since="", onto="", branch="from-base")
    eq_(ds.repo.get_active_branch(), "from-base")
    ok_(all(r["state"] == "clean"
            for r in ds.diff(fr=DEFAULT_BRANCH, to="from-base")))
    eq_(ds.repo.get_merge_base(["static", "from-base"]),
        ds.repo.get_hexsha("static^"))

    # We abort when an explicitly specified `onto` doesn't exist.
    ds.repo.checkout(DEFAULT_BRANCH)
    assert_result_count(
        ds.rerun(since="", onto="doesnotexist", branch="from-base",
                 on_failure="ignore"),
        1, status="error", action="run")


@known_failure_windows
@with_tempfile(mkdir=True)
def test_rerun_chain(path=None):
    ds = Dataset(path).create()
    commits = []

    with swallow_outputs():
        ds.run(f'echo x$({cat_command} grows) > grows')
    ds.repo.tag("first-run", commit=DEFAULT_BRANCH)

    for _ in range(3):
        commits.append(ds.repo.get_hexsha(DEFAULT_BRANCH))
        ds.rerun()
        _, info = get_run_info(ds, last_commit_msg(ds.repo))
        eq_(info["chain"], commits)

    ds.rerun(revision="first-run")
    _, info = get_run_info(ds, last_commit_msg(ds.repo))
    eq_(info["chain"], commits[:1])


@with_tempfile(mkdir=True)
def test_rerun_just_one_commit(path=None):
    ds = Dataset(path).create()
    if ds.repo.is_managed_branch():
        assert_status('impossible',
                      ds.rerun(branch="triggers-abort", on_failure="ignore"))
        raise SkipTest("Test incompatible with adjusted branch")

    ds.repo.checkout("orph", options=["--orphan"])
    ds.repo.call_git(["reset", "--hard"])
    ds.repo.config.reload()

    ds.run('echo static-content > static')
    eq_(len(ds.repo.get_revisions("HEAD")), 1)
    assert_raises(IncompleteResultsError, ds.rerun)
    assert_raises(IncompleteResultsError, ds.rerun, since="", onto="")

    # --script propagates the error.
    with swallow_outputs():
        assert_raises(IncompleteResultsError,
                      ds.rerun, since="", onto="", script="-")
    # --dry-run propagates the error.
    assert_raises(IncompleteResultsError,
                  ds.rerun, since="", onto="",
                  report=True, return_type="list")


@with_tempfile(mkdir=True)
def test_run_failure(path=None):
    ds = Dataset(path).create()
    subds = ds.create("sub")

    hexsha_initial = ds.repo.get_hexsha()

    if on_windows:
        # this does not do exactly the same as the cmd on other systems
        # but is close enough to make running the test worthwhile
        cmd_failing = "echo x>{} & false".format(op.join("sub", "grows"))
    else:
        cmd_failing = ("echo x$(cat {0}) > {0} && false"
                       .format(op.join("sub", "grows")))

    with assert_raises(IncompleteResultsError):
        ds.run(cmd_failing, result_renderer=None)
    eq_(hexsha_initial, ds.repo.get_hexsha())
    ok_(ds.repo.dirty)

    msgfile = op.join(path, ds.repo.get_git_dir(ds.repo), "COMMIT_EDITMSG")
    ok_exists(msgfile)

    ds.save(recursive=True, message_file=msgfile)
    assert_repo_status(ds.path)
    neq_(hexsha_initial, ds.repo.get_hexsha())

    outfile = op.join(subds.path, "grows")
    eq_('x \n' if on_windows else 'x\n', open(outfile).read())

    if on_windows:
        # FIXME: Make the remaining code compatible with Windows.
        return

    # There is no CommandError on rerun if the non-zero error matches the
    # original code.
    ds.rerun()
    eq_('xx\n', open(outfile).read())

    # On the other hand, we fail if we rerun a command and there is a non-zero
    # error that doesn't match.
    ds.run("[ ! -e bar ] && echo c >bar")
    assert_repo_status(ds.path)
    assert_in_results(ds.rerun(result_renderer=None, on_failure="ignore"),
                      action="run", status="error")

    # We don't show instructions if the caller specified us not to save.
    remove(msgfile)
    with assert_raises(IncompleteResultsError):
        ds.run("false", explicit=True, outputs=None, on_failure="stop")
    assert_false(op.exists(msgfile))


@with_tempfile(mkdir=True)
def test_rerun_branch(path=None):
    ds = Dataset(path).create()
    if ds.repo.is_managed_branch():
        assert_status('impossible',
                      ds.rerun(branch="triggers-abort", on_failure="ignore"))
        raise SkipTest("Test incompatible with adjusted branch")

    ds.repo.tag("prerun")

    outfile = op.join(path, "run-file")

    with swallow_outputs():
        ds.run(f'echo x$({cat_command} run-file) > run-file')
    ds.rerun()
    eq_('xx\n', open(outfile).read())

    with open(op.join(path, "nonrun-file"), "w") as f:
        f.write("foo")
    ds.save("nonrun-file")

    # Rerun the commands on a new branch that starts at the parent
    # commit of the first run.
    with swallow_outputs():
        ds.rerun(since="prerun", onto="prerun", branch="rerun")

    eq_(ds.repo.get_active_branch(), "rerun")
    eq_('xx\n', open(outfile).read())

    # NOTE: This test depends on the non-run commit above following a run
    # commit.  Otherwise, all the metadata (e.g., author date) aside from the
    # parent commit that is used to generate the commit ID may be set when
    # running the tests, which would result in two commits rather than three.
    for revrange in ["rerun.." + DEFAULT_BRANCH, DEFAULT_BRANCH + "..rerun"]:
        eq_(len(ds.repo.get_revisions(revrange)), 3)
    eq_(ds.repo.get_merge_base([DEFAULT_BRANCH, "rerun"]),
        ds.repo.get_hexsha("prerun"))

    # Start rerun branch at tip of current branch.
    ds.repo.checkout(DEFAULT_BRANCH)
    ds.rerun(since="prerun", branch="rerun2")
    eq_(ds.repo.get_active_branch(), "rerun2")
    eq_('xxxx\n', open(outfile).read())

    eq_(len(ds.repo.get_revisions(DEFAULT_BRANCH + "..rerun2")), 2)
    eq_(len(ds.repo.get_revisions("rerun2.." + DEFAULT_BRANCH)), 0)

    # Using an existing branch name fails.
    ds.repo.checkout(DEFAULT_BRANCH)
    assert_raises(IncompleteResultsError,
                  ds.rerun, since="prerun", branch="rerun2")


@skip_if_adjusted_branch
@with_tempfile(mkdir=True)
def test_rerun_cherry_pick(path=None):
    ds = Dataset(path).create()

    ds.repo.tag("prerun")
    ds.run('echo abc > runfile')
    with open(op.join(path, "nonrun-file"), "w") as f:
        f.write("foo")
    ds.save("nonrun-file")

    for onto, action in [("HEAD", "skip"), ("prerun", "pick")]:
        results = ds.rerun(since="prerun", onto=onto)
        assert_in_results(results, status='ok', rerun_action=action)


@skip_if_adjusted_branch
@with_tempfile(mkdir=True)
def test_rerun_invalid_merge_run_commit(path=None):
    ds = Dataset(path).create()
    ds.run("echo foo >>foo")
    ds.run("echo invalid >>invalid")
    run_msg = last_commit_msg(ds.repo)
    run_hexsha = ds.repo.get_hexsha()
    ds.repo.call_git(["reset", "--hard", DEFAULT_BRANCH + "~"])
    with open(op.join(ds.path, "non-run"), "w") as nrfh:
        nrfh.write("non-run")
    ds.save()
    # Assign two parents to the invalid run commit.
    commit = ds.repo.call_git_oneline(
        ["commit-tree", run_hexsha + "^{tree}", "-m", run_msg,
         "-p", run_hexsha + "^",
         "-p", ds.repo.get_hexsha()])
    ds.repo.call_git(["reset", "--hard", commit])
    hexsha_orig = ds.repo.get_hexsha()
    with swallow_logs(new_level=logging.WARN) as cml:
        ds.rerun(since="")
        assert_in("has run information but is a merge commit", cml.out)
    eq_(len(ds.repo.get_revisions(hexsha_orig + ".." + DEFAULT_BRANCH)), 1)


@with_tempfile(mkdir=True)
def test_rerun_outofdate_tree(path=None):
    ds = Dataset(path).create()
    input_file = op.join(path, "foo")
    output_file = op.join(path, "out")
    with open(input_file, "w") as f:
        f.write("abc\ndef")
    ds.save("foo", to_git=True)
    # Create initial run.
    ds.run(f'{grep_command} def foo > out')
    assert_in('def', open(output_file).read())
    # Change tree so that it is no longer compatible.
    ds.remove("foo")
    # Now rerunning should fail because foo no longer exists.
    assert_in_results(
        ds.rerun(revision=DEFAULT_BRANCH + "~",
                 result_renderer=None, on_failure="ignore"),
        status="error", action="run")


@with_tempfile(mkdir=True)
def test_rerun_ambiguous_revision_file(path=None):
    ds = Dataset(path).create()
    ds.run('echo ambig > ambig')
    ds.repo.tag("ambig", commit=DEFAULT_BRANCH)
    # Don't fail when "ambig" refers to both a file and revision.
    ds.rerun(since="", revision="ambig")
    eq_(len(ds.repo.get_revisions(DEFAULT_BRANCH)),
        len(ds.repo.get_revisions("ambig")))


@with_tree(tree={"subdir": {}})
def test_rerun_subdir(path=None):
    # Note: Using with_tree rather than with_tempfile is matters. The latter
    # calls realpath on the path, which masks a failure in the
    # TMPDIR="/var/tmp/sym link" test case
    ds = Dataset(path).create(force=True)
    subdir = op.join(path, 'subdir')
    with chpwd(subdir):
        run(touch_command + "test.dat")
    assert_repo_status(ds.path)

    # FIXME: A plain ok_file_under_git call doesn't properly resolve the file
    # in the TMPDIR="/var/tmp/sym link" test case. Temporarily call realpath.
    def ok_file_under_git_kludge(path, basename):
        ok_file_under_git(op.join(str(Path(path).resolve()), basename), annexed=True)

    ok_file_under_git_kludge(subdir, "test.dat")

    rec_msg, runinfo = get_run_info(ds, last_commit_msg(ds.repo))
    eq_(runinfo['pwd'], 'subdir')
    # now, rerun within root of the dataset
    with chpwd(ds.path):
        ds.rerun()
    assert_repo_status(ds.path)
    ok_file_under_git_kludge(subdir, "test.dat")
    # and not on top
    assert_raises(AssertionError, ok_file_under_git,
                  op.join(ds.path, "test.dat"), annexed=True)

    # but if we run ds.run -- runs within top of the dataset
    with chpwd(subdir):
        ds.run(touch_command + "test2.dat")
    assert_repo_status(ds.path)
    ok_file_under_git_kludge(ds.path, "test2.dat")
    rec_msg, runinfo = get_run_info(ds, last_commit_msg(ds.repo))
    eq_(runinfo['pwd'], '.')
    # now, rerun within subdir -- smoke for now
    with chpwd(subdir):
        ds.rerun()


@with_tree(tree={"d": {"to_modify": "content1"},
                 "to_remove": "content2",
                 "to_modify": "content3",
                 "unchanged": "content4"})
def test_new_or_modified(path=None):
    def get_new_or_modified(*args, **kwargs):
        return [op.relpath(ap["path"], path)
                for ap in new_or_modified(diff_revision(*args, **kwargs))]

    ds = Dataset(path).create(force=True, annex=False)

    # Check out an orphan branch so that we can test the "one commit
    # in a repo" case.
    ds.repo.checkout("orph", options=["--orphan"])
    ds.save()
    assert_false(ds.repo.dirty)
    eq_(len(ds.repo.get_revisions("HEAD")), 1)
    # Diffing doesn't fail when the branch contains a single commit.
    assert_in("to_modify", get_new_or_modified(ds, "HEAD"))

    # New files are detected, deletions are not.
    ds.repo.remove(["to_remove"])
    ok_(ds.repo.dirty)

    with open(op.join(path, "to_add"), "w") as f:
        f.write("content5")
    ds.repo.add(["to_add"])
    ds.repo.commit("add one, remove another")

    eq_(get_new_or_modified(ds, "HEAD"),
        ["to_add"])

    # Modifications are detected.
    with open(op.join(path, "to_modify"), "w") as f:
        f.write("updated 1")
    with open(op.join(path, "d/to_modify"), "w") as f:
        f.write("updated 2")
    ds.save(["to_modify", "d/to_modify"])

    eq_(set(get_new_or_modified(ds, "HEAD")),
        {"to_modify", op.join("d", "to_modify")})

    # Non-HEAD revisions work.
    ds.repo.commit("empty", options=["--allow-empty"])
    assert_false(get_new_or_modified(ds, "HEAD"))
    eq_(set(get_new_or_modified(ds, "HEAD~")),
        {"to_modify", op.join("d", "to_modify")})


@with_tempfile(mkdir=True)
def test_rerun_script(path=None):
    ds = Dataset(path).create()
    ds.run("echo a >foo")
    ds.run([touch_command + "bar"], message='BAR', sidecar=True)
    # a run record sidecar file was added with the last commit
    assert(any(d['path'].startswith(op.join(ds.path, '.datalad', 'runinfo'))
               for d in ds.rerun(report=True, return_type='item-or-list')['diff']))
    bar_hexsha = ds.repo.get_hexsha(DEFAULT_BRANCH)

    script_file = op.join(path, "commands.sh")

    ds.rerun(script=script_file)
    ok_exists(script_file)
    with open(script_file) as sf:
        lines = sf.readlines()
        assert_in(touch_command + "bar\n", lines)
        # The commit message is there too.
        assert_in("# BAR\n", lines)
        assert_in("# (record: {})\n".format(bar_hexsha), lines)
        assert_not_in("echo a >foo\n", lines)

    ds.rerun(since="", script=script_file)
    with open(script_file) as sf:
        lines = sf.readlines()
        assert_in(touch_command + "bar\n", lines)
        # Automatic commit messages aren't included.
        assert_not_in("# echo a >foo\n", lines)
        assert_in("echo a >foo\n", lines)

    # --script=- writes to stdout.
    with patch("sys.stdout", new_callable=StringIO) as cmout:
        ds.rerun(script="-")
        assert_in(touch_command + "bar",
                  cmout.getvalue().splitlines())


@slow  # ~10s
@known_failure_windows
@pytest.mark.xfail(reason="push fails due to IncompleteResultsError "
                          "[remote rejected] (branch is currently checked out)")
# ^ Issue only happens on appveyor, Python itself implodes. Cannot be
#   reproduced on a real win7 box
# Comment above looks outdated. Last trial on Appveyor failed, but seems related
# to unresolved globs:
# https://ci.appveyor.com/project/mih/datalad/builds/37951288/job/yxx47i3vtola2wek
@with_tree(tree={"input.dat": "input",
                 "extra-input.dat": "extra input",
                 "s0": {"s1_0": {"s2": {"a.dat": "a",
                                        "b.txt": "b"}},
                        "s1_1": {"s2": {"c.dat": "c",
                                        "d.txt": "d"}},
                        "ss": {"e.dat": "e"}}})
@with_tempfile(mkdir=True)
def test_run_inputs_outputs(src=None, path=None):
    for subds in [("s0", "s1_0", "s2"),
                  ("s0", "s1_1", "s2"),
                  ("s0", "s1_0"),
                  ("s0", "s1_1"),
                  ("s0", "ss"),
                  ("s0",)]:
        Dataset(op.join(*((src,) + subds))).create(force=True).save()
    src_ds = Dataset(src).create(force=True)
    src_ds.save()

    ds = install(path, source=src,
                 result_xfm='datasets', return_type='item-or-list')
    assert_false(ds.repo.file_has_content("input.dat"))
    assert_false(ds.repo.file_has_content("extra-input.dat"))

    # The specified inputs and extra inputs will be retrieved before the run.
    # (Use run_command() to access the extra_inputs argument.)
    list(run_command("{} {{inputs}} {{inputs}} >doubled.dat"
                     .format('type' if on_windows else 'cat'),
                     dataset=ds,
                     inputs=["input.dat"], extra_inputs=["extra-input.dat"]))

    assert_repo_status(ds.path)
    ok_(ds.repo.file_has_content("input.dat"))
    ok_(ds.repo.file_has_content("extra-input.dat"))
    ok_(ds.repo.file_has_content("doubled.dat"))
    with open(op.join(path, "doubled.dat")) as fh:
        content = fh.read()
        assert_in("input", content)
        assert_not_in("extra-input", content)

    # Rerunning the commit will also get the input file.
    ds.repo.drop(["input.dat", "extra-input.dat"], options=["--force"])
    assert_false(ds.repo.file_has_content("input.dat"))
    assert_false(ds.repo.file_has_content("extra-input.dat"))
    ds.rerun()
    ok_(ds.repo.file_has_content("input.dat"))
    ok_(ds.repo.file_has_content("extra-input.dat"))

    # Test different combinations of globs and explicit files.
    inputs = ["a.dat", "b.dat", "c.txt", "d.txt"]
    create_tree(ds.path, {i: i for i in inputs})

    ds.save()
    ds.push(inputs, to=DEFAULT_REMOTE)
    ds.repo.drop(inputs, options=["--force"])

    test_cases = [(["*.dat"], ["a.dat", "b.dat"]),
                  (["*.dat", "c.txt"], ["a.dat", "b.dat", "c.txt"]),
                  (["*"], inputs)]

    for idx, (inputs_arg, expected_present) in enumerate(test_cases):
        assert_false(any(ds.repo.file_has_content(i) for i in inputs))

        ds.run("cd .> dummy{}".format(idx), inputs=inputs_arg)
        ok_(all(ds.repo.file_has_content(f) for f in expected_present))
        # Globs are stored unexpanded by default.
        assert_in(inputs_arg[0], last_commit_msg(ds.repo))
        ds.repo.drop(inputs, options=["--force"])

    # --input can be passed a subdirectory.
    create_tree(ds.path, {"subdir": {"a": "subdir a",
                                     "b": "subdir b"}})
    ds.save("subdir")
    ds.push(["subdir/a", "subdir/b"], to=DEFAULT_REMOTE)
    ds.repo.drop("subdir", options=["--force"])
    ds.run("cd .> subdir-dummy", inputs=[op.join(ds.path, "subdir")])
    ok_(all(ds.repo.file_has_content(op.join("subdir", f)) for f in ["a", "b"]))

    # Inputs are specified relative to a dataset's subdirectory.
    ds.repo.drop(op.join("subdir", "a"), options=["--force"])
    with chpwd(op.join(path, "subdir")):
        run("cd .> subdir-dummy1", inputs=["a"])
    ok_(ds.repo.file_has_content(op.join("subdir", "a")))

    # --input=. runs "datalad get ."
    ds.run("cd .> dot-dummy", inputs=["."])
    eq_(ds.repo.get_annexed_files(),
        ds.repo.get_annexed_files(with_content_only=True))
    # On rerun, we get all files, even those that weren't in the tree at the
    # time of the run.
    create_tree(ds.path, {"after-dot-run": "after-dot-run content"})
    ds.save()
    ds.push(["after-dot-run"], to=DEFAULT_REMOTE)
    ds.repo.drop(["after-dot-run"], options=["--force"])
    ds.rerun(DEFAULT_BRANCH + "^")
    ds.repo.file_has_content("after-dot-run")

    # --output will unlock files that are present.
    ds.repo.get("a.dat")
    ds.run("echo ' appended' >>a.dat", outputs=["a.dat"])
    with open(op.join(path, "a.dat")) as fh:
        eq_(fh.read(), "a.dat' appended' \n" if on_windows else "a.dat appended\n" )

    # --output will remove files that are not present.
    ds.repo.drop(["a.dat", "d.txt"], options=["--force"])
    ds.run("echo ' appended' >>a.dat", outputs=["a.dat"])
    if not on_windows:
        # MIH doesn't yet understand how to port this
        with open(op.join(path, "a.dat")) as fh:
            eq_(fh.read(), " appended\n" )

    # --input can be combined with --output.
    ds.run("echo ' appended' >>a.dat", inputs=["a.dat"], outputs=["a.dat"])
    if not on_windows:
        # MIH doesn't yet understand how to port this
        with open(op.join(path, "a.dat")) as fh:
            eq_(fh.read(), " appended\n appended\n")

    ds.create('sub')
    ds.run("echo sub_orig >sub/subfile")
    ds.run("echo sub_overwrite >sub/subfile", outputs=["sub/subfile"])
    ds.drop("sub/subfile", reckless='kill')
    ds.run("echo sub_overwrite >sub/subfile", outputs=["sub/subfile"])

    # --input/--output globs can be stored in expanded form.
    ds.run("cd .> expand-dummy", inputs=["a.*"], outputs=["b.*"], expand="both")
    assert_in("a.dat", last_commit_msg(ds.repo))
    assert_in("b.dat", last_commit_msg(ds.repo))

    res = ds.rerun(report=True, return_type='item-or-list')
    eq_(res["run_info"]['inputs'], ["a.dat"])
    eq_(res["run_info"]['outputs'], ["b.dat"])

    # We uninstall subdatasets to fully resolve globs.
    ds.drop("s0", what='all', reckless='kill', recursive=True)
    assert_false(Dataset(op.join(path, "s0")).is_installed())
    ds.run("echo {inputs} >globbed-subds", inputs=["s0/s1_*/s2/*.dat"])
    ok_file_has_content(
        op.join(ds.path, "globbed-subds"),
        "'s0\\s1_0\\s2\\a.dat' 's0\\s1_1\\s2\\c.dat'" if on_windows
        else "s0/s1_0/s2/a.dat s0/s1_1/s2/c.dat",
        strip=True)

    ds_ss = Dataset(op.join(path, "s0", "ss"))
    assert_false(ds_ss.is_installed())
    ds.run("echo blah >{outputs}", outputs=["s0/ss/out"])
    ok_(ds_ss.is_installed())
    ok_file_has_content(op.join(ds.path, "s0", "ss", "out"),
                        "blah",
                        strip=True)


@with_tree({"foo": "foo"})
def test_run_inputs_no_annex_repo(path=None):
    ds = Dataset(path).create(annex=False, force=True)
    ds.save()
    # Running --input in a plain Git repo doesn't fail.
    ds.run("cd .> dummy", inputs=["*"])
    ok_exists(op.join(ds.path, "dummy"))
    ds.rerun()


@skip_if_adjusted_branch
@with_tree(tree={"to_modify": "to_modify"})
def test_rerun_explicit(path=None):
    ds = Dataset(path).create(force=True)

    ds.run("echo o >> foo", explicit=True, outputs=["foo"])
    with open(op.join(ds.path, "foo")) as ifh:
        orig_content = ifh.read()
        orig_head = ds.repo.get_hexsha(DEFAULT_BRANCH)

    # Explicit rerun is allowed in a dirty tree.
    ok_(ds.repo.dirty)
    ds.rerun(explicit=True)
    eq_(orig_head, ds.repo.get_hexsha(DEFAULT_BRANCH + "~1"))
    with open(op.join(ds.path, "foo")) as ifh:
        eq_(orig_content * 2, ifh.read())

    # --since also works.
    ds.rerun(since="", explicit=True)
    eq_(orig_head,
        # Added two rerun commits.
        ds.repo.get_hexsha(DEFAULT_BRANCH + "~3"))

    # With just untracked changes, we can rerun with --onto.
    ds.rerun(since="", onto="", explicit=True)
    eq_(ds.repo.get_hexsha(orig_head + "^"),
        # Reran the four run commits from above on the initial base.
        ds.repo.get_hexsha("HEAD~4"))

    # But checking out a new HEAD can fail when there are modifications.
    ds.repo.checkout(DEFAULT_BRANCH)
    ok_(ds.repo.dirty)
    ds.repo.add(["to_modify"], git=True)
    ds.save()
    assert_false(ds.repo.dirty)
    with open(op.join(ds.path, "to_modify"), "a") as ofh:
        ofh.write("more")
    ok_(ds.repo.dirty)

    with assert_raises(CommandError):
        ds.rerun(onto="", since="", explicit=True)


@with_tempfile(mkdir=True)
def test_rerun_assume_ready(path=None):
    ds = Dataset(path).create()
    repo = ds.repo
    (repo.pathobj / "f1").write_text("f1\n")
    ds.save()

    def double_in_both_cmd(src, dest1, dest2):
        return [
            sys.executable, "-c",
            "import sys; import os; import os.path as op; "
            "content = open(sys.argv[-3]).read() * 2; "
            "d1 = sys.argv[-2]; d2 = sys.argv[-1]; "
            "op.lexists(d1) and os.unlink(d1); "
            "op.lexists(d2) and os.unlink(d2); "
            "open(d1, 'w').write(content); open(d2, 'w').write(content)",
            src, dest1, dest2]

    ds.run(double_in_both_cmd("f1", "out1", "out2"), outputs=["out1"])
    # Drop the content so that we remove instead of unlock, making the test is
    # more meaningful on an adjusted branch.
    ds.drop(["out1", "out2"], reckless='kill')
    # --assume-ready affects both explicitly specified and automatic outputs.
    res = ds.rerun(assume_ready="outputs")
    assert_not_in_results(res, action="remove")


# underlying code cannot deal with adjusted branches
# https://github.com/datalad/datalad/pull/3817
@known_failure_windows
@with_tree(tree={"a.in": "a", "b.in": "b", "c.out": "c",
                 "subdir": {}})
def test_placeholders(path=None):
    ds = Dataset(path).create(force=True)
    ds.save()
    assert_repo_status(ds.path)
    # ATTN windows is sensitive to spaces before redirect symbol
    ds.run("echo {inputs}>{outputs}", inputs=[".", "*.in"], outputs=["c.out"])
    ok_file_has_content(op.join(path, "c.out"), "a.in b.in\n")

    hexsha_before = ds.repo.get_hexsha()
    ds.rerun()
    eq_(hexsha_before, ds.repo.get_hexsha())

    # ATTN windows is sensitive to spaces before redirect symbol
    ds.run("echo {inputs[0]}>getitem", inputs=["*.in"])
    ok_file_has_content(op.join(path, "getitem"), "a.in\n")

    ds.run("echo {pwd} >expanded-pwd")
    ok_file_has_content(op.join(path, "expanded-pwd"), path,
                        strip=True)

    ds.run("echo {dspath} >expanded-dspath")
    ok_file_has_content(op.join(path, "expanded-dspath"), ds.path,
                        strip=True)

    subdir_path = op.join(path, "subdir")
    with chpwd(subdir_path):
        run("echo {pwd} >expanded-pwd")
    ok_file_has_content(op.join(path, "subdir", "expanded-pwd"), subdir_path,
                        strip=True)
    eq_(get_run_info(ds, last_commit_msg(ds.repo))[1]["pwd"],
        "subdir")

    # Double brackets can be used to escape placeholders.
    ds.run("cd .> {{inputs}}", inputs=["*.in"])
    ok_exists(op.join(path, "{inputs}"))

    # rerun --script expands the placeholders.
    with patch("sys.stdout", new_callable=StringIO) as cmout:
        ds.rerun(script="-", since="")
        script_out = cmout.getvalue()
        assert_in("echo a.in b.in>c.out", script_out)
        assert_in("echo {} >expanded-pwd".format(subdir_path),
                  script_out)
        assert_in("echo {} >expanded-dspath".format(ds.path),
                  script_out)

    assert_result_count(
        ds.run("{unknown_placeholder}", on_failure="ignore"),
        1, status="impossible", action="run")

    # Configured placeholders.
    ds.config.add("datalad.run.substitutions.license", "gpl3", scope="local")
    ds.run("echo {license} >configured-license")
    ok_file_has_content(op.join(path, "configured-license"), "gpl3",
                        strip=True)
    # --script handles configured placeholders.
    with patch("sys.stdout", new_callable=StringIO) as cmout:
        ds.rerun(script="-")
        assert_in("gpl3", cmout.getvalue())

    ds.run("echo {tmpdir} >tout")
    ok_file_has_content(op.join(path, "tout"), ".*datalad-run.*", re_=True)


def test_rerun_commit_message_check():
    assert_raises(ValueError,
                  get_run_info,
                  None,
                  """\
[DATALAD RUNCMD] no command

=== Do not change lines below ===
{
 "pwd": ".",
 "exit": 0
}
^^^ Do not change lines above ^^^""")

    assert_raises(ValueError,
                  get_run_info,
                  None,
                  """\
[DATALAD RUNCMD] junk json

=== Do not change lines below ===
{
 "pwd": ".,
 "cmd": "echo ok >okfile",
 "exit": 0
}
^^^ Do not change lines above ^^^""")

    subject, info = get_run_info(
        None,
        """\
[DATALAD RUNCMD] fine

=== Do not change lines below ===
{
 "pwd": ".",
 "cmd": "echo ok >okfile",
 "exit": 0
}
^^^ Do not change lines above ^^^""")
    eq_(subject, "fine")
    assert_dict_equal(info,
                      {"pwd": ".", "cmd": "echo ok >okfile", "exit": 0})



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import os
from pathlib import PurePosixPath

from ...tests.utils_pytest import (
    SkipTest,
    assert_raises,
    eq_,
    with_tempfile,
)
from ...utils import (
    chpwd,
    on_windows,
    rmtree,
)
from ..path import (
    abspath,
    curdir,
    get_parent_paths,
    get_filtered_paths_,
    robust_abspath,
    split_ext,
)

import pytest


@with_tempfile(mkdir=True)
def test_robust_abspath(tdir=None):
    with chpwd(tdir):
        eq_(robust_abspath(curdir), tdir)
        try:
            if os.environ.get('DATALAD_ASSERT_NO_OPEN_FILES'):
                raise Exception("cannot test under such pressure")
            rmtree(tdir)
        except Exception as exc:
            # probably windows or above exception
            raise SkipTest(
                "Cannot test in current environment") from exc

        assert_raises(OSError, abspath, curdir)
        eq_(robust_abspath(curdir), tdir)


def test_split_ext():
    eq_(split_ext("file"), ("file", ""))

    eq_(split_ext("file.py"), ("file", ".py"))
    eq_(split_ext("file.tar.gz"), ("file", ".tar.gz"))
    eq_(split_ext("file.toolong.gz"), ("file.toolong", ".gz"))

    eq_(split_ext("file.a.b.c.d"), ("file", ".a.b.c.d"))
    eq_(split_ext("file.a.b.cccc.d"), ("file", ".a.b.cccc.d"))
    eq_(split_ext("file.a.b.ccccc.d"), ("file.a.b.ccccc", ".d"))

    eq_(split_ext("file.a.b..c"), ("file", ".a.b..c"))


@pytest.mark.parametrize("sep", [None, '/', '\\'])
def test_get_parent_paths(sep):
    if sep is None:
        gpp = get_parent_paths
    else:
        from functools import partial
        gpp = partial(get_parent_paths, sep=sep)

    # sanity/border checks
    eq_(gpp([], []), [])
    eq_(gpp([], ['a']), [])
    eq_(gpp(['a'], ['a']), ['a'])

    # Helper to provide testing across different seps and platforms while
    # specifying only POSIX paths here in the test
    def _p(path):
        if sep is None:
            return path
        else:
            return path.replace('/', sep)
    _pp = lambda paths: list(map(_p, paths))

    # no absolute paths anywhere
    if on_windows:
        assert_raises(ValueError, gpp, 'C:\\a', ['a'])
        assert_raises(ValueError, gpp, ['a'], 'C:\\a')
    elif sep != '\\':  # \ does not make it absolute
        assert_raises(ValueError, gpp, _p('/a'), ['a'])
        assert_raises(ValueError, gpp, ['a'], [_p('/a')])
    assert_raises(ValueError, gpp, [_p('a//a')], ['a'])
    # dups the actual code but there is no other way AFAIK
    asep = {'/': '\\', None: '\\', '\\': '/'}[sep]
    assert_raises(ValueError, gpp, [f'a{asep}a'], ['a'])

    paths = _pp(['a', 'a/b', 'a/b/file', 'c', 'd/sub/123'])

    eq_(gpp(paths, []), paths)
    eq_(gpp(paths, [], True), [])

    # actually a tricky one!  we should check in descending lengths etc
    eq_(gpp(paths, paths), paths)
    # every path is also its own parent
    eq_(gpp(paths, paths, True), paths)

    # subdatasets not for every path -- multiple paths hitting the same parent,
    # and we will be getting only a single entry
    # to mimic how git ls-tree operates
    eq_(gpp(paths, ['a']), ['a', 'c', _p('d/sub/123')])
    eq_(gpp(paths, ['a'], True), ['a'])

    # and we get the deepest parent
    eq_(gpp(_pp(['a/b/file', 'a/b/file2']), _pp(['a', 'a/b'])), _pp(['a/b']))


def test_get_filtered_paths_():
    # just to avoid typing all the same
    def gfp(*args, **kwargs):
        return list(get_filtered_paths_(*args, **kwargs))

    assert gfp(['a', 'b'], ['a', 'c']) == ['a']
    assert gfp(['a', 'b'], ['b']) == ['b']
    assert gfp(['a', 'b'], ['c']) == []

    assert gfp(['a', 'b'], ['a/b', 'c']) == []  # a is not subpath of a/b
    assert gfp(['a', 'b'], ['a/b', 'c'], include_within_path=True) == ['a']  # a is not subpath of a/b

    # all paths returned due to '.', and order is sorted
    paths = ['a', 'b', '1/2/3', 'abc']
    paths_sorted = sorted(paths)
    assert gfp(paths, ['.']) == paths_sorted
    assert gfp(paths, paths_sorted) == paths_sorted
    assert gfp(paths, paths_sorted, include_within_path=True) == paths_sorted
    # we can take a mix of str and Path
    assert gfp([PurePosixPath(paths[0])] + paths[1:], ['.']) == paths_sorted


    # nothing within empty "filter_paths" matches -- so no paths yielded
    assert gfp(paths, []) == []

    assert_raises(ValueError, gfp, ['/a'], [])
    assert_raises(ValueError, gfp, [PurePosixPath('/a')], [])
    assert_raises(ValueError, gfp, ['a'], ['/a'])
    assert_raises(ValueError, gfp, ['../a'], ['a'])


# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Base classes to custom git-annex remotes (e.g. extraction from archives)"""

from __future__ import absolute_import

__docformat__ = 'restructuredtext'


import logging
from collections import Counter

lgr = logging.getLogger('datalad.customremotes')

from annexremote import (
    RemoteError,
    UnsupportedRequest,
)
from datalad.customremotes import SpecialRemote

URI_PREFIX = "dl"


class AnnexCustomRemote(SpecialRemote):
    # default properties
    COST = 100
    AVAILABILITY = "local"

    def __init__(self, annex):  # , availability=DEFAULT_AVAILABILITY):
        super().__init__(annex)
        # TODO self.info = {}, self.configs = {}

        # OPT: a counter to increment upon successful encounter of the scheme
        # (ATM only in gen_URLS but later could also be used in other
        # requests).  This would allow to consider schemes in order of
        # decreasing success instead of arbitrary hardcoded order
        self._scheme_hits = Counter({s: 0 for s in self.SUPPORTED_SCHEMES})

    @classmethod
    def _get_custom_scheme(cls, prefix):
        """Helper to generate custom datalad URL prefixes
        """
        # prefix which will be used in all URLs supported by this custom remote
        # https://tools.ietf.org/html/rfc2718 dictates "URL Schemes" standard
        # 2.1.2 suggests that we do use // since all of our URLs will define
        # some hierarchical structure.  But actually since we might encode
        # additional information (such as size) into the URL, it will not be
        # strictly conforming it. Thus we will not use //
        return "%s+%s" % (URI_PREFIX, prefix)  # if .PREFIX else '')

    # Helper methods
    def gen_URLS(self, key):
        """Yield URL(s) associated with a key, and keep stats on protocols."""
        nurls = 0
        for scheme, _ in self._scheme_hits.most_common():
            scheme_ = scheme + ":"
            scheme_urls = self.annex.geturls(key, scheme_)
            if scheme_urls:
                # note: generator would cease to exist thus not asking
                # for URLs for other schemes if this scheme is good enough
                self._scheme_hits[scheme] += 1
                for url in scheme_urls:
                    nurls += 1
                    yield url
        self.annex.debug("Processed %d URL(s) for key %s", nurls, key)

    # Protocol implementation
    def initremote(self):
        pass

    def prepare(self):
        pass

    def transfer_store(self, key, local_file):
        raise UnsupportedRequest('This special remote cannot store content')

    def remove(self, key):
        raise RemoteError("Removal of content from urls is not possible")

    def getcost(self):
        return self.COST

    def getavailability(self):
        return self.AVAILABILITY


# this function only has anecdotal value and is not used anywhere
def generate_uuids():
    """Generate UUIDs for our remotes. Even though quick, for
    consistency pre-generated and recorded in consts.py"""
    import uuid
    return {
        remote: str(uuid.uuid5(
            uuid.NAMESPACE_URL,
            'http://datalad.org/specialremotes/%s' % remote))
        for remote in {'datalad', 'datalad-archives'}
    }


def init_datalad_remote(repo, remote, encryption=None, autoenable=False,
                        opts=[]):
    """Initialize datalad special remote"""
    from datalad.consts import DATALAD_SPECIAL_REMOTES_UUIDS
    lgr.info("Initializing special remote %s", remote)
    remote_opts = [
        'encryption=%s' % str(encryption).lower(),
        'type=external',
        'autoenable=%s' % str(bool(autoenable)).lower(),
        'externaltype=%s' % remote
    ]
    # use unique uuid for our remotes
    # This should help with merges of disconnected repos etc
    # ATM only datalad/datalad-archives is expected,
    # so on purpose getitem
    remote_opts.append('uuid=%s' % DATALAD_SPECIAL_REMOTES_UUIDS[remote])
    return repo.init_remote(remote, remote_opts + opts)


def ensure_datalad_remote(repo, remote=None,
                          encryption=None, autoenable=False):
    """Initialize and enable datalad special remote if it isn't already.

    Parameters
    ----------
    repo : AnnexRepo
    remote : str, optional
        Special remote name. This should be one of the values in
        datalad.consts.DATALAD_SPECIAL_REMOTES_UUIDS and defaults to
        datalad.consts.DATALAD_SPECIAL_REMOTE.
    encryption, autoenable : optional
        Passed to `init_datalad_remote`.
    """
    from datalad.consts import (
        DATALAD_SPECIAL_REMOTE,
        DATALAD_SPECIAL_REMOTES_UUIDS,
    )

    remote = remote or DATALAD_SPECIAL_REMOTE

    uuid = DATALAD_SPECIAL_REMOTES_UUIDS.get(remote)
    if not uuid:
        raise ValueError("'{}' is not a known datalad special remote: {}"
                         .format(remote,
                                 ", ".join(DATALAD_SPECIAL_REMOTES_UUIDS)))
    name = repo.get_special_remotes().get(uuid, {}).get("name")

    if not name:
        init_datalad_remote(repo, remote,
                            encryption=encryption, autoenable=autoenable)
    elif repo.is_special_annex_remote(name, check_if_known=False):
        lgr.debug("datalad special remote '%s' is already enabled", name)
    else:
        lgr.info("datalad special remote '%s' found. Enabling", name)
        repo.enable_remote(name)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support of custom remotes (e.g. extraction from archives)

"""

__docformat__ = 'restructuredtext'

__all__ = ['RemoteError', 'SpecialRemote']

from annexremote import (
    ProtocolError,
    SpecialRemote as _SpecialRemote,
    RemoteError as _RemoteError,
)
from datalad.support.exceptions import format_exception_with_cause
from datalad.ui import ui


class RemoteError(_RemoteError):
    def __str__(self):
        # this is a message given to remote error, if any
        exc_str = super().__str__()
        # this is the cause ala `raise from`
        exc_cause = getattr(self, '__cause__', None)
        if exc_cause:
            # if we have a cause, collect the cause all the way down
            # we can do quite some chaining
            exc_cause = format_exception_with_cause(exc_cause)
        if exc_str and exc_cause:
            # with have the full picture
            msg = f'{exc_str} -caused by- {exc_cause}'
        elif exc_str and not exc_cause:
            # only a custom message
            msg = exc_str
        elif not exc_str and exc_cause:
            # only the cause
            msg = exc_cause
        else:
            # nothing, shame!
            msg = 'exception with unknown cause'
        # prevent multiline messages, they would be swallowed
        # or kill the protocol
        return msg.replace('\n', '\\n')


class SpecialRemote(_SpecialRemote):
    """Common base class for all of DataLad's special remote implementations"""

    def __init__(self, annex):
        super(SpecialRemote, self).__init__(annex=annex)
        # instruct annex backend UI to use this remote
        if ui.backend == 'annex':
            ui.set_specialremote(self)

    def message(self, msg, type='debug'):
        handler = dict(
            debug=self.annex.debug,
            info=self.annex.info,
            error=self.annex.error,
        ).get(type, self.annex.debug)

        # ensure that no multiline messages are sent, they would cause a
        # protocol error
        msg = msg.replace('\n', '\\n')

        try:
            handler(msg)
        except ProtocolError:
            # INFO not supported by annex version.
            # If we can't have an actual info message, at least have a
            # debug message.
            self.annex.debug(msg)

    def send_progress(self, progress):
        """Indicates the current progress of the transfer (in bytes).

        May be repeated any number of times during the transfer process.

        Too frequent updates are wasteful but bear in mind that this is used
        both to display a progress meter for the user, and for
        ``annex.stalldetection``. So, sending an update on each 1% of the file
        may not be frequent enough, as it could appear to be a stall when
        transferring a large file.

        Parameters
        ----------
        progress : int
            The current progress of the transfer in bytes.
        """
        # This method is called by AnnexSpecialRemoteProgressBar through an
        # obscure process that involves multiple layers of abstractions for
        # UIs, providers, downloaders, progressbars, which is only happening
        # within the environment of a running special remote process though
        # a combination of circumstances.
        #
        # The main purpose of this method is to have a place to leave this
        # comment within the code base of the special remotes, in order to
        # aid future souls having to sort this out.
        # (and to avoid having complex code make direct calls to internals
        # of this class, making things even more complex)
        self.annex.progress(progress)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Tooling for creating a publication target on GitHub-like systems
"""

import logging
import re
from urllib.parse import (
    urljoin,
    urlparse,
)

import requests

from datalad import cfg as dlcfg
from datalad.distribution.dataset import (
    EnsureDataset,
    require_dataset,
)
from datalad.downloaders.credentials import Token
from datalad.downloaders.http import DEFAULT_USER_AGENT
from datalad.interface.common_opts import (
    publish_depends,
    recursion_flag,
    recursion_limit,
)
from datalad.interface.results import get_status_dict
from datalad.support.constraints import (
    EnsureChoice,
    EnsureNone,
    EnsureStr,
)
from datalad.support.param import Parameter
from datalad.ui import ui
from datalad.utils import todo_interface_for_extensions


lgr = logging.getLogger('datalad.distributed.create_sibling_ghlike')


class _GitHubLike(object):
    """Helper class with a platform abstraction for GitHub-like services
    """
    # (short) lower-case name of the target platform
    name = None
    # (longer) name with fancy capitalization
    fullname = None
    # all API endpoint without base URL!
    # to create a repo in an organization
    create_org_repo_endpoint = None
    # to create a repo under the authenticated user
    create_user_repo_endpoint = None
    # query for props of the authenticated users
    get_authenticated_user_endpoint = None
    # query for repository properties
    get_repo_info_endpoint = None
    # HTTP response codes for particular events
    # repo created successfully
    response_code_repo_created = requests.codes.created
    # auth failure
    response_code_unauthorized = requests.codes.forbidden

    # extra config settings to be used for a remote pointing to the
    # target platform
    extra_remote_settings = {}

    # to be used (in modified form) by create_sibling_*() commands that
    # utilize this platform abstraction
    create_sibling_params = dict(
        dataset=Parameter(
            args=("--dataset", "-d",),
            doc="""dataset to create the publication target for. If not given,
            an attempt is made to identify the dataset based on the current
            working directory""",
            constraints=EnsureDataset() | EnsureNone()),
        reponame=Parameter(
            args=('reponame',),
            metavar='[<org-name>/]<repo-(base)name>',
            doc="""repository name, optionally including an '<organization>/'
            prefix if the repository shall not reside under a user's namespace.
            When operating recursively, a suffix will be appended to this name
            for each subdataset""",
            constraints=EnsureStr()),
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
        name=Parameter(
            args=('-s', '--name',),
            metavar='NAME',
            doc="""name of the sibling in the local dataset installation
            (remote name)""",
            constraints=EnsureStr() | EnsureNone()),
        existing=Parameter(
            args=("--existing",),
            constraints=EnsureChoice(
                'skip', 'error', 'reconfigure'),
            doc="""behavior when already existing or configured
            siblings are discovered: skip the dataset ('skip'), update the
            configuration ('reconfigure'), or fail ('error').""",),
        credential=Parameter(
            args=('--credential',),
            constraints=EnsureStr() | EnsureNone(),
            metavar='NAME',
            doc="""name of the credential providing a personal access token
            to be used for authorization. The token can be supplied via
            configuration setting 'datalad.credential.<name>.token', or
            environment variable DATALAD_CREDENTIAL_<NAME>_TOKEN, or will
            be queried from the active credential store using the provided
            name. If none is provided, the host-part of the API URL is used
            as a name (e.g. 'https://api.github.com' -> 'api.github.com')"""),
        api=Parameter(
            args=('--api',),
            constraints=EnsureStr() | EnsureNone(),
            metavar='URL',
            # TODO consider default instance via config
            doc="""API endpoint of the Git hosting service instance"""),
        access_protocol=Parameter(
            args=("--access-protocol",),
            constraints=EnsureChoice('https', 'ssh', 'https-ssh'),
            doc="""access protocol/URL to configure for the sibling. With
            'https-ssh' SSH will be used for write access, whereas HTTPS
            is used for read access."""),
        description=Parameter(
            args=("--description",),
            doc="""Brief description, displayed on the project's page""",
            constraints=EnsureStr() | EnsureNone()),
        publish_depends=publish_depends,
        private=Parameter(
            args=("--private",),
            action="store_true",
            default=False,
            doc="""if set, create a private repository"""),
        dry_run=Parameter(
            args=("--dry-run",),
            action="store_true",
            doc="""if set, no repository will be created, only tests for
            sibling name collisions will be performed, and would-be repository
            names are reported for all relevant datasets"""),
    )

    def __init__(self, url, credential, require_token=True, token_info=None):
        if self.name is None or self.fullname is None:
            raise NotImplementedError(
                'GitHub-like platform must have name and fullname properties')
        if not url:
            raise ValueError(f'API URL required for {self.fullname}')

        self.api_url = url
        self._user_info = None
        self._set_request_headers(
            credential,
            f'An access token is required for {url}' \
            + f'. {token_info}' if token_info else '',
            require_token,
        )
        self._set_extra_remote_settings()

    def _set_extra_remote_settings(self):
        target_name = urlparse(self.api_url).netloc
        config_section = "datalad.create-sibling-ghlike.extra-remote-settings.{}".format(target_name)
        target_specific_settings = {
            option: dlcfg.get_value(config_section, option)
            for option in dlcfg.options(config_section)
        }
        self.extra_remote_settings = {
            **self.extra_remote_settings,
            **target_specific_settings,
        }

    @todo_interface_for_extensions
    def _set_request_headers(self, credential_name, auth_info, require_token):
        if credential_name is None:
            credential_name = urlparse(self.api_url).netloc

        try:
            self.auth = Token(credential_name)(
                instructions=auth_info)['token']
        except Exception as e:
            lgr.debug('Token retrieval failed: %s', e)
            lgr.warning(
                'Cannot determine authorization token for %s', credential_name)
            if require_token:
                raise ValueError(
                    f'Authorization required for {self.fullname}, '
                    f'cannot find token for a credential {credential_name}.')
            else:
                lgr.warning("No token found for credential '%s'", credential_name)
                self.auth = 'NO-TOKEN-AVAILABLE'

        self.request_headers = {
            'user-agent': DEFAULT_USER_AGENT,
            'authorization': f'token {self.auth}',
        }

    @property
    def authenticated_user(self):
        """Lazy query/reporting of properties for the authenticated user

        Returns
        -------
        dict
        """
        if self._user_info:
            return self._user_info

        endpoint = urljoin(self.api_url, self.get_authenticated_user_endpoint)
        headers = self.request_headers
        r = requests.get(endpoint, headers=headers)
        # make sure any error-like situation causes noise
        r.raise_for_status()
        self._user_info = r.json()
        return self._user_info

    # TODO what are the actual constraints?
    def normalize_reponame(self, path):
        """Turn name into a GitHub-like service compliant repository name

        Useful for sanitizing directory names.
        """
        return re.sub(r'\s+', '_', re.sub(r'[/\\]+', '-', path))

    def get_dataset_reponame_mapping(
            self, ds, name, reponame, existing, recursive, recursion_limit,
            res_kwargs):
        """Discover all relevant datasets locally, and build remote repo names
        """
        dss = _get_present_datasets(ds, recursive, recursion_limit)
        # check for existing remote configuration
        toprocess = []
        toyield = []
        for d in dss:
            if existing not in ('reconfigure', 'replace') and \
                    name in d.repo.get_remotes():
                toyield.append(get_status_dict(
                    ds=d,
                    status='error' if existing == 'error' else 'notneeded',
                    message=('already has a configured sibling "%s"', name),
                    **res_kwargs)
                )
                continue
            gh_reponame = reponame if d == ds else \
                '{}-{}'.format(
                    reponame,
                    self.normalize_reponame(
                        str(d.pathobj.relative_to(ds.pathobj))))
            toprocess.append((d, gh_reponame))
        return toprocess, toyield

    def get_siblingname(self, siblingname):
        """Generate a (default) sibling name, if none is given

        Returns
        -------
        str
        """
        if siblingname:
            return siblingname

        if self.api_url:
            siblingname = urlparse(self.api_url).netloc

        if not siblingname:
            raise ValueError(
                'No valid sibling name given or determined: {}'.format(
                    siblingname))

        return siblingname

    def create_repo(self, ds, reponame, organization, private, dry_run,
                    description, existing):
        """Create a repository on the target platform

        Returns
        -------
        dict
          Result record, with status='ok' when all is good, status='error'
          when unrecoverably broken, status='impossible' when recoverably
          broken

        Raises
        ------
        Exception
          Any unhandled condition (in particular unexpected non-success
          HTTP response codes) will raise an exception.
        """
        res = self.repo_create_request(
            reponame, organization, private, dry_run, description)

        if res.get('status') == 'impossible' and res.get('preexisted'):
            # we cannot create, because there is something in the target
            # spot
            orguser = organization or self.authenticated_user['login']

            if existing == 'reconfigure':
                # we want to use the existing one instead
                # query properties, report, and be done
                repo_props = self.repo_get_request(orguser, reponame)
                res.update(
                    status='notneeded',
                    # return in full
                    host_response=repo_props,
                    # perform some normalization
                    **self.normalize_repo_properties(repo_props)
                )
            elif existing == 'replace':
                # only implemented for backward compat with
                # create-sibling-github
                _msg = ('repository "%s" already exists', reponame)
                if ui.is_interactive:
                    remove = ui.yesno(
                        "Do you really want to remove it?",
                        title=_msg[0] % _msg[1],
                        default=False
                    )
                else:
                    return dict(
                        res,
                        status='impossible',
                        message=(
                            _msg[0] + " Remove it manually first or "
                            "rerun DataLad in an interactive shell "
                            "to confirm this action.",
                            _msg[1]),
                    )
                if not remove:
                    return dict(
                        res,
                        status='impossible',
                        message=_msg,
                    )
                # remove the beast in cold blood
                self.repo_delete_request(
                    organization or self.authenticated_user['login'],
                    reponame)
                # try creating now
                return self.create_repo(
                    ds, reponame, organization, private, dry_run,
                    description, existing)

        # TODO intermediate error handling?

        return res

    def repo_get_request(self, orguser, reponame):
        """Perform request to query for repo properties

        Returns
        -------
        dict
          The JSON payload of the response.
        """
        # query information on the existing repo and use that
        # to complete the task
        r = requests.get(
            urljoin(
                self.api_url,
                self.get_repo_info_endpoint.format(
                    user=orguser,
                    repo=reponame)),
            headers=self.request_headers,
        )
        # make sure any error-like situation causes noise
        r.raise_for_status()
        return r.json()

    def repo_delete_request(self, orguser, reponame):
        """Perform request to delete a named repo on the platform

        Must be implemented in subclasses for particular target platforms.
        """
        raise NotImplementedError

    def create_repos(self, dsrepo_map, siblingname, organization,
                     private, dry_run, description, res_kwargs,
                     existing, access_protocol,
                     publish_depends):
        """Create a series of repos on the target platform

        This method handles common conditions in a uniform platform-agnostic
        fashion, and sets local sibling configurations for created/located
        repositories.

        Yields
        ------
        dict
          Result record
        """
        for d, reponame in dsrepo_map:
            res = self.create_repo(
                d, reponame, organization, private, dry_run, description,
                existing)
            # blend reported results with standard properties
            res = dict(
                res,
                **res_kwargs)

            if res.get('preexisted') and existing == 'skip':
                # we came here, despite initial checking for conflicting
                # sibling names. this means we found an unrelated repo
                res['status'] = 'error'
                res['message'] = (
                    "A repository '%s' already exists at '%s', "
                    "use existing=reconfigure to use it as a sibling",
                    reponame, self.api_url)

            if 'message' not in res:
                if not dry_run:
                    res['message'] = (
                        "sibling repository '%s' created at %s",
                        siblingname, res.get('html_url')
                    )
                else:
                    # can't know url when request was not made
                    res['message'] = (
                        "would create sibling '%s' and repository '%s%s'",
                        siblingname,
                        organization + "/" if organization else "",
                        reponame
                    )
            # report to caller
            yield get_status_dict(**res)

            if res['status'] not in ('ok', 'notneeded'):
                # something went wrong, do not proceed
                continue

            if dry_run:
                continue

            if res['status'] == 'notneeded' \
                    and existing not in ('reconfigure', 'replace'):
                # nothing to do anymore, when no reconfiguration is desired
                continue

            # lastly configure the local datasets
            for var_name, var_value in \
                    self.extra_remote_settings.items():
                var = 'remote.{}.{}'.format(siblingname, var_name)
                if existing in ('reconfigure', 'replace'):
                    d.config.set(var, var_value, scope='local')
                elif var not in d.config:
                    d.config.add(var, var_value, scope='local')
            yield from d.siblings(
                'configure',
                name=siblingname,
                url=res['ssh_url']
                if access_protocol == 'ssh'
                else res['clone_url'],
                pushurl=res['ssh_url']
                if access_protocol == 'https-ssh' else None,
                recursive=False,
                # TODO fetch=True, maybe only if one existed already
                publish_depends=publish_depends,
                return_type='generator',
                result_renderer='disabled')

    def repo_create_request(self, reponame, organization, private,
                            dry_run=False, description=None):
        """Perform a request to create a repo on the target platform

        Also implements reporting of "fake" results in dry-run mode.

        Returns
        -------
        dict
          Result record, but see repo_create_response() for details.
        """
        endpoint = urljoin(
            self.api_url,
            self.create_org_repo_endpoint.format(
                organization=organization)
            if organization else
            self.create_user_repo_endpoint)
        desc_text = description if description is not None else 'some default'
        data = {
            'name': reponame,
            'description': desc_text,
            'private': private,
            'auto_init': False,
        }
        headers = self.request_headers

        if dry_run:
            return dict(
                status='ok',
                request_url=endpoint,
                request_data=data,
                request_headers=headers,
            )
        r = requests.post(
            endpoint,
            json=data,
            headers=headers,
        )
        return self.repo_create_response(r)

    def repo_create_response(self, r):
        """Handling of repo creation request responses

        Normalizes error handling and reporting.

        Returns
        -------
        dict
          Result record

        Raises
        ------
        Exception
          Raises for any unhandled HTTP error response code.
        """
        try:
            response = r.json()
        except Exception as e:
            lgr.debug('Cannot get JSON payload of %s [%s]' , r, e)
            response = {}
        lgr.debug('%s responded with %s %s', self.fullname, r, response)
        if r.status_code == self.response_code_repo_created:
            return dict(
                status='ok',
                preexisted=False,
                # return in full
                host_response=response,
                # perform some normalization
                **self.normalize_repo_properties(response)
            )
        elif r.status_code == requests.codes.unprocessable and \
                'already exist' in response.get('message', ''):
            return dict(
                status='impossible',
                message='repository already exists',
                preexisted=True,
            )
        elif r.status_code in (self.response_code_unauthorized,
                               requests.codes.forbidden):
            return dict(
                status='error',
                message=('unauthorized: %s', response.get('message')),
            )
        elif r.status_code == requests.codes.internal_server_error:
            return dict(
                status='error',
                message=response.get('message', '').strip() or 'Server returned error code %d without any further information' % requests.codes.internal_server_error,
            )
        # make sure any error-like situation causes noise
        r.raise_for_status()
        # catch-all
        raise RuntimeError(f'Unexpected host response: {response}')

    def normalize_repo_properties(self, response):
        """Normalize the essential response properties for the result record

        Importantly, `clone_url` is a URL that DataLad can directly clone
        from, and that should be the default preferred access method for read
        access by the largest possible audience. Critically, a particular
        platform, might advertise SSH as default, but DataLad might promote
        anonymous HTTP-access as a default, if supported.

        Returns
        -------
        dict
        """
        return dict(
            reponame=response.get('name'),
            private=response.get('private'),
            clone_url=response.get('clone_url'),
            ssh_url=response.get('ssh_url'),
            html_url=response.get('html_url'),
        )


def _create_sibling(
        platform,
        reponame,
        dataset=None,
        recursive=False,
        recursion_limit=None,
        name=None,
        existing='error',
        access_protocol='https',
        publish_depends=None,
        private=False,
        description=None,
        dry_run=False):
    """Helper function to conduct sibling creation on a target platform

    Parameters match the respective create_sibling_*() commands.
    `platform` is an instance of a subclass of `_GitHubLike`.

    Yields
    ------
    dict
      Result record.
    """

    orgname, reponame = split_org_repo(reponame)

    # apply whatever normalization or default selection
    name = platform.get_siblingname(name)

    lgr.debug("Repository organization name: '%s'", orgname)

    if reponame != platform.normalize_reponame(reponame):
        raise ValueError(
            f'Invalid name for a {platform.fullname} project: {reponame}')

    lgr.debug("Repository basename: '%s'", reponame)

    # what to operate on
    ds = require_dataset(
        dataset,
        check_installed=True,
        purpose=f'create {platform.fullname} sibling(s)')

    res_kwargs = dict(
        action=\
        f'create_sibling_{platform.name} [dry-run]'
        if dry_run else
        f'create_sibling_{platform.name}',
        logger=lgr,
        refds=ds.path,
    )

    toprocess, filterresults = platform.get_dataset_reponame_mapping(
        ds, name, reponame, existing, recursive, recursion_limit,
        res_kwargs
    )
    yield from filterresults

    if not toprocess:
        # all skipped
        return

    lgr.debug("Will process %i dataset(s)", len(toprocess))

    yield from platform.create_repos(
        toprocess,
        name,
        orgname,
        private,
        dry_run,
        description,
        res_kwargs,
        existing,
        access_protocol,
        publish_depends,
    )


def split_org_repo(name):
    """Split a potential organization name prefix from a repo's full name

    Returns
    -------
    (None, reponame) or (orgname, reponame)
    """
    split = name.split('/', maxsplit=1)
    if len(split) < 2:
        return None, name
    else:
        return split[0], split[1]


def _get_present_datasets(ds, recursive, recursion_limit):
    """Return list of (sub)dataset instances for all locally present datasets
    """
    # gather datasets and essential info
    # dataset instance and mountpoint relative to the top
    toprocess = [ds]
    if recursive:
        for sub in ds.subdatasets(
                # we want to report on missing dataset in here
                state='any',
                recursive=recursive,
                recursion_limit=recursion_limit,
                result_xfm='datasets',
                result_renderer='disabled',
                return_type='generator'):
            if not sub.is_installed():
                lgr.info('Ignoring unavailable subdataset %s', sub)
                continue
            toprocess.append(sub)
    return toprocess



# emacs: -*- mode: python-mode; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test functioning of the datalad main cmdline utility """

import os
import re
from io import StringIO
from unittest.mock import patch

import pytest

import datalad
from datalad import __version__
from datalad.api import (
    Dataset,
    create,
)
from datalad.cmd import StdOutErrCapture
from datalad.cmd import WitlessRunner as Runner
from datalad.interface.base import get_interface_groups
from datalad.tests.utils_pytest import (
    SkipTest,
    assert_equal,
    assert_in,
    assert_not_in,
    assert_raises,
    assert_re_in,
    eq_,
    in_,
    ok_,
    ok_startswith,
    on_windows,
    skip_if_no_module,
    slow,
    with_tempfile,
)
from datalad.ui.utils import (
    get_console_width,
    get_terminal_size,
)
from datalad.utils import chpwd

from ..helpers import get_commands_from_groups
from ..main import main


def run_main(args, exit_code=0, expect_stderr=False):
    """Run main() of the datalad, do basic checks and provide outputs

    Parameters
    ----------
    args : list
        List of string cmdline arguments to pass
    exit_code : int
        Expected exit code. Would raise AssertionError if differs
    expect_stderr : bool or string
        Whether to expect stderr output. If string -- match

    Returns
    -------
    stdout, stderr  strings
       Output produced
    """
    was_mode = datalad.__api
    try:
        # we need to catch "stdout" from multiple places:
        # sys.stdout but also from the UI, which insists on holding
        # a dedicated handle
        fakeout = StringIO()
        fakeerr = StringIO()
        with patch('sys.stderr', new=fakeerr) as cmerr, \
             patch('sys.stdout', new=fakeout) as cmout, \
             patch.object(datalad.ui.ui._ui, 'out', new=fakeout):
            with assert_raises(SystemExit) as cm:
                main(["datalad"] + list(args))
            eq_('cmdline', datalad.get_apimode())
            assert_equal(cm.value.code, exit_code)
            stdout = cmout.getvalue()
            stderr = cmerr.getvalue()
            if expect_stderr is False:
                assert_equal(stderr, "")
            elif expect_stderr is True:
                # do nothing -- just return
                pass
            else:
                # must be a string
                assert_equal(stderr, expect_stderr)
    finally:
        # restore what we had
        datalad.__api = was_mode

    return stdout, stderr


def get_all_commands() -> list:
    return list(get_commands_from_groups(get_interface_groups()))


def assert_all_commands_present(out):
    """Helper to reuse to assert that all known commands are present in output
    """
    for cmd in get_all_commands():
        assert_re_in(fr"\b{cmd}\b", out, match=False)


# TODO: switch to stdout for --version output
def test_version():
    # we just get a version if not asking for a version of some command
    stdout, stderr = run_main(['--version'], expect_stderr=True)
    eq_(stdout.rstrip(), "datalad %s" % datalad.__version__)

    stdout, stderr = run_main(['clone', '--version'], expect_stderr=True)
    ok_startswith(stdout, 'datalad %s\n' % datalad.__version__)
    # since https://github.com/datalad/datalad/pull/2733 no license in --version
    assert_not_in("Copyright", stdout)
    assert_not_in("Permission is hereby granted", stdout)

    try:
        import datalad_container
    except ImportError:
        pass  # not installed, cannot test with extension
    else:
        stdout, stderr = run_main(['containers-list', '--version'], expect_stderr=True)
        eq_(stdout, 'datalad_container %s\n' % datalad_container.__version__)


def test_help_np():
    stdout, stderr = run_main(['--help-np'])

    # Let's extract section titles:
    # enough of bin/datalad and .tox/py27/bin/datalad -- guarantee consistency! ;)
    ok_startswith(stdout, 'Usage: datalad')
    # Sections start/end with * if ran under DATALAD_HELP2MAN mode
    sections = [l[1:-1] for l in filter(re.compile(r'^\*.*\*$').match, stdout.split('\n'))]
    for s in {'Essential',
              'Miscellaneous',
              'General information',
              'Global options',
              'Plumbing',
              }:
        assert_in(s, sections)
        # should be present only one time!
        eq_(stdout.count(f'*{s}*'), 1)

    # check that we have global options actually listed after "Global options"
    # ATM -c is the first such option
    assert re.search(r"Global options\W*-c ", stdout, flags=re.MULTILINE)
    # and -c should be listed only once - i.e. that we do not duplicate sections
    # and our USAGE summary has only [global-opts]
    assert re.match(r"Usage: .*datalad.* \[global-opts\] command \[command-opts\]", stdout)
    assert stdout.count(' -c ') == 1

    assert_all_commands_present(stdout)

    if not get_terminal_size()[0] or 0:
        raise SkipTest(
            "Could not determine terminal size, skipping the rest of the test")

    # none of the lines must be longer than 80 chars
    # TODO: decide on   create-sibling and possibly
    # rewrite-urls
    accepted_width = get_console_width()

    long_lines = ["%d %s" % (len(l), l) for l in stdout.split('\n')
                  if len(l) > accepted_width and
                  '{' not in l  # on nd70 summary line is unsplit
                  ]
    if long_lines:
        raise AssertionError(
            "Following lines in --help output were longer than %s chars:\n%s"
            % (accepted_width, '\n'.join(long_lines))
        )


def test_dashh():
    stdout, stderr = run_main(['-h'])
    # Note: for -h we do not do ad-hoc tune up of Usage: to guarantee having
    # datalad instead of python -m nose etc, so we can only verify that we have
    # options listed
    assert_re_in(r'^Usage: .*\[', stdout.splitlines()[0])
    assert_all_commands_present(stdout)
    assert_re_in('Use .--help. to get more comprehensive information', stdout.splitlines())


def test_dashh_clone():
    # test -h on a sample command
    stdout, stderr = run_main(['clone', '-h'])
    assert_re_in(r'^Usage: .* clone \[', stdout.splitlines()[0])
    assert_re_in('Use .--help. to get more comprehensive information', stdout.splitlines())


def test_usage_on_insufficient_args():
    stdout, stderr = run_main(['install'], exit_code=2, expect_stderr=True)
    ok_startswith(stderr, 'usage:')


def test_subcmd_usage_on_unknown_args():
    stdout, stderr = run_main(['get', '--murks'], exit_code=1, expect_stderr=True)
    in_('get', stdout)


def test_combined_short_option():
    stdout, stderr = run_main(['-fjson'], exit_code=2, expect_stderr=True)
    assert_not_in("unrecognized argument", stderr)
    assert_in("too few arguments", stderr)


# https://github.com/datalad/datalad/issues/6814
@with_tempfile(mkdir=True)
def test_conflicting_short_option(tempdir=None):
    # datalad -f|--format   requires a value. regression made parser ignore command
    # and its options
    with chpwd(tempdir):  # can't just use -C tempdir since we do "in process" run_main
        run_main(['create', '-f'])


# apparently a bit different if following a good one so let's do both
err_invalid = "error: (invalid|too few arguments|unrecognized argument)"
err_insufficient = err_invalid  # "specify"


@pytest.mark.parametrize(
    "opts,err_str",
    [
        (('--buga',), err_invalid),
        (('--dbg', '--buga'), err_invalid),
        (('--dbg',), err_insufficient),
        (tuple(), err_insufficient),
    ]
)
def test_incorrect_option(opts, err_str):
    # The first line used to be:
    # stdout, stderr = run_main((sys.argv[0],) + opts, expect_stderr=True, exit_code=2)
    # But: what do we expect to be in sys.argv[0] here?
    # It depends on how we invoke the test.
    # - nosetests -s -v datalad/cmdline/tests/test_main.py would result in:
    #   sys.argv[0}=='nosetests'
    # - python -m nose -s -v datalad/cmdline/tests/test_main.py would result in:
    #   sys.argv[0}=='python -m nose'
    # - python -c "import nose; nose.main()" -s -v datalad/cmdline/tests/test_main.py would result in:
    #   sys.argv[0]=='-c'
    # This led to failure in case sys.argv[0] contained an option, that was
    # defined to be a datalad option too, therefore was a 'known_arg' and was
    # checked to meet its constraints.
    # But sys.argv[0] actually isn't used by main at all. It simply doesn't
    # matter what's in there. The only thing important to pass here is `opts`.
    stdout, stderr = run_main(opts, expect_stderr=True, exit_code=2)
    out = stdout + stderr
    assert_in("usage: ", out)
    assert_re_in(err_str, out, match=False)


@pytest.mark.parametrize(
    "script",
    [
        'datalad',
        'git-annex-remote-datalad-archives',
        'git-annex-remote-datalad',
    ]
)
def test_script_shims(script):
    runner = Runner()
    if not on_windows:

        from shutil import which
        which(script)

    # and let's check that it is our script
    out = runner.run([script, '--version'], protocol=StdOutErrCapture)
    version = out['stdout'].rstrip()
    mod, version = version.split(' ', 1)
    assert_equal(mod, 'datalad')
    # we can get git and non git .dev version... so for now
    # relax
    get_numeric_portion = lambda v: [x for x in re.split('[+.]', v) if x.isdigit()]
    # extract numeric portion
    assert get_numeric_portion(version), f"Got no numeric portion from {version}"
    assert_equal(get_numeric_portion(__version__),
                 get_numeric_portion(version))


@with_tempfile(mkdir=True)
def test_cfg_override(path=None):
    with chpwd(path):
        # use 'wtf' to dump the config
        # should be rewritten to use `configuration`
        cmd = ['datalad', 'wtf', '-S', 'configuration', '-s', 'some']
        # control
        out = Runner().run(cmd, protocol=StdOutErrCapture)['stdout']
        assert_not_in('datalad.dummy: this', out)
        # ensure that this is not a dataset's cfg manager
        assert_not_in('datalad.dataset.id', out)
        # env var
        out = Runner(env=dict(os.environ, DATALAD_DUMMY='this')).run(
            cmd, protocol=StdOutErrCapture)['stdout']
        assert_in('datalad.dummy: this', out)
        # cmdline arg
        out = Runner().run([cmd[0], '-c', 'datalad.dummy=this'] + cmd[1:],
                           protocol=StdOutErrCapture)['stdout']
        assert_in('datalad.dummy: this', out)

        # now create a dataset in the path. the wtf plugin will switch to
        # using the dataset's config manager, which must inherit the overrides
        create(dataset=path, annex=False)
        # control
        out = Runner().run(cmd, protocol=StdOutErrCapture)['stdout']
        assert_not_in('datalad.dummy: this', out)
        # ensure that this is a dataset's cfg manager
        assert_in('datalad.dataset.id', out)
        # env var
        out = Runner(env=dict(os.environ, DATALAD_DUMMY='this')).run(
            cmd, protocol=StdOutErrCapture)['stdout']
        assert_in('datalad.dummy: this', out)
        # cmdline arg
        out = Runner().run([cmd[0], '-c', 'datalad.dummy=this'] + cmd[1:],
                           protocol=StdOutErrCapture)['stdout']
        assert_in('datalad.dummy: this', out)

        # set a config
        run_main([
            'configuration', '--scope', 'local', 'set', 'mike.item=some'])
        # verify it is successfully set
        assert 'some' == run_main([
            'configuration', 'get', 'mike.item'])[0].strip()
        # verify that an override can unset the config
        # we cannot use run_main(), because the "singleton" instance of the
        # dataset we are in is still around in this session, and with it
        # also its config managers that we will not be able to post-hoc
        # overwrite with this method. Instead, we'll execute in a subprocess.
        assert '' == Runner().run([
            'datalad', '-c', ':mike.item',
            'configuration', 'get', 'mike.item'],
            protocol=StdOutErrCapture)['stdout'].strip()
        # verify the effect is not permanent
        assert 'some' == Runner().run([
            'datalad',
            'configuration', 'get', 'mike.item'],
            protocol=StdOutErrCapture)['stdout'].strip()


def test_incorrect_cfg_override():
    run_main(['-c', 'some', 'wtf'], exit_code=3)
    run_main(['-c', 'some=', 'wtf'], exit_code=3)
    run_main(['-c', 'some.var', 'wtf'], exit_code=3)
    run_main(['-c', 'some.var=', 'wtf'], exit_code=3)


@with_tempfile
def test_librarymode(path=None):
    Dataset(path).create()
    was_mode = datalad.__runtime_mode
    try:
        # clean --dry-run is just a no-op command that is cheap
        # to execute. It has no particular role here, other than
        # to make the code pass the location where library mode
        # should be turned on via the cmdline API
        run_main(['-c', 'datalad.runtime.librarymode=yes', 'clean',
                  '-d', path, '--dry-run'])
        ok_(datalad.in_librarymode())
    finally:
        # restore pre-test behavior
        datalad.__runtime_mode = was_mode
        datalad.cfg.overrides.pop('datalad.runtime.librarymode')


@with_tempfile
def test_completion(out_fn=None):
    skip_if_no_module('argcomplete')

    from datalad.cmd import WitlessRunner
    runner = WitlessRunner()

    def get_completions(s: str, expected) -> list:
        """Run 'datalad' external command and collect completions

        Parameters
        ----------
        s: str
          what to append to 'datalad ' invocation
        expected: iterable of str
          What entries to expect - would raise AssertionError if any is
          not present in output
        exit_code: int, optional
          If incomplete/malformed we seems to get 2, most frequently used
          so default

        Returns
        -------
        list of str
          Entries output
        """
        if os.path.exists(out_fn):  # reuse but ensure it is gone
            os.unlink(out_fn)
        comp_line = f'datalad {s}'
        runner.run(
            comp_line.split(' '),
            env=dict(os.environ,
                     _ARGCOMPLETE='1',
                     _ARGCOMPLETE_STDOUT_FILENAME=out_fn,
                     COMP_LINE=comp_line,
                     # without -1 seems to get "finished completion", someone can investigate more
                     COMP_POINT=str(len(comp_line)-1),  # always at the end ATM
                     ))
        with open(out_fn, 'rb') as f:
            entries = f.read().split(b'\x0b')
            entries = [e.decode() for e in entries]
        diff = set(expected).difference(entries)
        if diff:
            raise AssertionError(
                f"Entries {sorted(diff)} were expected but not found in the completion output: {entries}"
            )
        return entries  # for extra analyzes if so desired

    all_commands = get_all_commands()
    get_completions('i', {'install'})
    get_completions(' ', ['--dbg', '-c'] + all_commands)
    # if command already matches -- we get only that hit ATM, not others which begin with it
    get_completions('create', ['create '])
    get_completions('create -', ['--dataset'])
    # but for incomplete one we do get all create* commands
    get_completions('creat', [c for c in all_commands if c.startswith('create')])



# -*- coding: utf-8 -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test status command"""

import os.path as op

import datalad.utils as ut
from datalad.api import status
from datalad.core.local.status import get_paths_by_ds
from datalad.distribution.dataset import Dataset
from datalad.support.annexrepo import AnnexRepo
from datalad.support.exceptions import (
    CommandError,
    IncompleteResultsError,
    NoDatasetFound,
)
from datalad.tests.utils_pytest import (
    OBSCURE_FILENAME,
    SkipTest,
    assert_dict_equal,
    assert_in,
    assert_in_results,
    assert_not_in_results,
    assert_raises,
    assert_repo_status,
    assert_result_count,
    assert_status,
    eq_,
    get_deeply_nested_structure,
    has_symlink_capability,
    with_tempfile,
)
from datalad.utils import (
    chpwd,
    on_windows,
)


@with_tempfile(mkdir=True)
def test_runnin_on_empty(path=None):
    # empty repo
    repo = AnnexRepo(path, create=True)
    # just wrap with a dataset
    ds = Dataset(path)
    # and run status ... should be good and do nothing
    eq_([], ds.status(result_renderer='disabled'))


@with_tempfile(mkdir=True)
@with_tempfile()
@with_tempfile(mkdir=True)
def test_status_basics(path=None, linkpath=None, otherdir=None):
    if has_symlink_capability():
        # make it more complicated by default
        ut.Path(linkpath).symlink_to(path, target_is_directory=True)
        path = linkpath

    with chpwd(path):
        assert_raises(NoDatasetFound, status)
    ds = Dataset(path).create()
    # outcome identical between ds= and auto-discovery
    with chpwd(path):
        assert_raises(IncompleteResultsError, status, path=otherdir)
        stat = status(result_renderer='disabled')
    eq_(stat, ds.status(result_renderer='disabled'))
    assert_status('ok', stat)
    # we have a bunch of reports (be vague to be robust to future changes
    assert len(stat) > 2
    # check the composition
    for s in stat:
        eq_(s['status'], 'ok')
        eq_(s['action'], 'status')
        eq_(s['state'], 'clean')
        eq_(s['type'], 'file')
        assert_in('gitshasum', s)
        assert_in('bytesize', s)
        eq_(s['refds'], ds.path)


@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_status_nods(path=None, otherpath=None):
    ds = Dataset(path).create()
    assert_result_count(
        ds.status(path=otherpath, on_failure='ignore', result_renderer='disabled'),
        1,
        status='error',
        message=('path not underneath the reference dataset %s', ds.path))
    otherds = Dataset(otherpath).create()
    assert_result_count(
        ds.status(path=otherpath, on_failure='ignore', result_renderer='disabled'),
        1,
        path=otherds.path,
        status='error',
        message=('path not underneath the reference dataset %s', ds.path))


@with_tempfile(mkdir=True)
@with_tempfile()
def test_status(_path=None, linkpath=None):
    # do the setup on the real path, not the symlink, to have its
    # bugs not affect this test of status()
    ds = get_deeply_nested_structure(str(_path))
    if has_symlink_capability():
        # make it more complicated by default
        ut.Path(linkpath).symlink_to(_path, target_is_directory=True)
        path = linkpath
    else:
        path = _path

    ds = Dataset(path)
    if has_symlink_capability():
        assert ds.pathobj != ds.repo.pathobj

    # spotcheck that annex status reporting and availability evaluation
    # works
    assert_result_count(
        ds.status(annex='all', result_renderer='disabled'),
        1,
        path=str(ds.pathobj / 'subdir' / 'annexed_file.txt'),
        key='MD5E-s5--275876e34cf609db118f3d84b799a790.txt',
        has_content=True,
        objloc=str(ds.repo.pathobj / '.git' / 'annex' / 'objects' /
        # hashdir is different on windows
        ('f33' if ds.repo.is_managed_branch() else '7p') /
        ('94b' if ds.repo.is_managed_branch() else 'gp') /
        'MD5E-s5--275876e34cf609db118f3d84b799a790.txt' /
        'MD5E-s5--275876e34cf609db118f3d84b799a790.txt'))

    plain_recursive = ds.status(recursive=True, result_renderer='disabled')
    # check integrity of individual reports
    for res in plain_recursive:
        # anything that is an "intended" symlink should be reported
        # as such
        if 'link2' in str(res['path']):
            assert res['type'] == 'symlink', res
        # every item must report its parent dataset
        assert_in('parentds', res)

    # bunch of smoke tests
    # query of '.' is same as no path
    eq_(plain_recursive, ds.status(path='.', recursive=True,
                                   result_renderer='disabled'))
    # duplicate paths do not change things
    eq_(plain_recursive, ds.status(path=['.', '.'], recursive=True,
                                   result_renderer='disabled'))
    # neither do nested paths
    eq_(plain_recursive,
        ds.status(path=['.', 'subds_modified'], recursive=True,
                  result_renderer='disabled'))
    # when invoked in a subdir of a dataset it still reports on the full thing
    # just like `git status`, as long as there are no paths specified
    with chpwd(op.join(path, 'directory_untracked')):
        plain_recursive = status(recursive=True, result_renderer='disabled')
    # should be able to take absolute paths and yield the same
    # output
    eq_(plain_recursive, ds.status(path=ds.path, recursive=True,
                                   result_renderer='disabled'))

    # query for a deeply nested path from the top, should just work with a
    # variety of approaches
    rpath = op.join('subds_modified', 'subds_lvl1_modified',
                    OBSCURE_FILENAME + u'_directory_untracked')
    apathobj = ds.pathobj / rpath
    apath = str(apathobj)
    # ds.repo.pathobj will have the symlink resolved
    arealpath = ds.repo.pathobj / rpath
    # TODO include explicit relative path in test
    for p in (rpath, apath, arealpath, None):
        if p is None:
            # change into the realpath of the dataset and
            # query with an explicit path
            with chpwd(ds.repo.path):
                res = ds.status(path=op.join('.', rpath), result_renderer='disabled')
        else:
            res = ds.status(path=p, result_renderer='disabled')
        assert_result_count(
            res,
            1,
            state='untracked',
            type='directory',
            refds=ds.path,
            # path always comes out a full path inside the queried dataset
            path=apath,
        )

    assert_result_count(
        ds.status(
            recursive=True, result_renderer='disabled'),
        1,
        path=apath)
    # limiting recursion will exclude this particular path
    assert_result_count(
        ds.status(
            recursive=True,
            recursion_limit=1, result_renderer='disabled'),
        0,
        path=apath)
    # negative limit is unlimited limit
    eq_(
        ds.status(recursive=True, recursion_limit=-1, result_renderer='disabled'),
        ds.status(recursive=True, result_renderer='disabled')
    )

    # check integrity of individual reports with a focus on how symlinks
    # are reported in annex-mode
    # this is different from plain git-mode, which reports types as-is
    # from the git record
    for res in ds.status(recursive=True, annex='basic',
                         result_renderer='disabled'):
        # anything that is an "intended" symlink should be reported
        # as such. In contrast, anything that is a symlink for mere
        # technical reasons (annex using it for something in some mode)
        # should be reported as the thing it is representing (i.e.
        # a file)
        if 'link2' in str(res['path']):
            assert res['type'] == 'symlink', res
        else:
            assert res['type'] != 'symlink', res

@with_tempfile(mkdir=True)
def test_untracked_annex_query(path=None):
    # test for #7032
    ds = Dataset(path).create()
    (ds.pathobj / 'untracked_file.txt').write_text(u'dummy')
    res = ds.status(annex='basic', path='untracked_file.txt')
    assert_not_in_results(
        res,
        error_message='File unknown to git',
        )


# https://github.com/datalad/datalad-revolution/issues/64
# breaks when the tempdir is a symlink
@with_tempfile(mkdir=True)
def test_subds_status(path=None):
    ds = Dataset(path).create()
    subds = ds.create('subds')
    assert_repo_status(ds.path)
    subds.create('someotherds')
    assert_repo_status(subds.path)
    assert_repo_status(ds.path, modified=['subds'])
    assert_result_count(
        ds.status(path='subds', result_renderer='disabled'),
        1,
        # must be modified, not added (ds was clean after it was added)
        state='modified',
        type='dataset',
        path=subds.path,
        refds=ds.path)

    # path="." gets treated as "this dataset's content" without requiring a
    # trailing "/"...
    assert_result_count(
        subds.status(path=".", result_renderer='disabled'),
        1,
        type="dataset",
        path=op.join(subds.path, "someotherds"),
        refds=subds.path)

    # ... and so does path=<path/to/ds>.
    assert_result_count(
        subds.status(path=subds.path, result_renderer='disabled'),
        1,
        type="dataset",
        path=op.join(subds.path, "someotherds"),
        refds=subds.path)

    assert_result_count(
        subds.status(path=op.join(subds.path, op.pardir, "subds"),
                     result_renderer='disabled'),
        1,
        type="dataset",
        path=op.join(subds.path, "someotherds"),
        refds=subds.path)

    assert_result_count(
        subds.status(path=op.join(subds.path, op.curdir),
                     result_renderer='disabled'),
        1,
        type="dataset",
        path=op.join(subds.path, "someotherds"),
        refds=subds.path)


@with_tempfile
def test_status_symlinked_dir_within_repo(path=None):
    if not has_symlink_capability():
        raise SkipTest("Can't create symlinks")
    # <path>
    # |-- bar -> <path>/foo
    # `-- foo
    #     `-- f
    ds = Dataset(path).create()
    foo = ds.pathobj / "foo"
    foo.mkdir()
    (foo / "f").write_text("content")
    (ds.pathobj / "bar").symlink_to(foo, target_is_directory=True)
    ds.save()
    bar_f = ds.pathobj / "bar" / "f"

    def call():
        return ds.status(path=[bar_f], annex="availability",
                         on_failure="ignore", result_renderer='disabled')

    if ds.repo.git_annex_version < "8.20200522" \
        or (on_windows and ds.repo.git_annex_version < "10.20220525"):
        # version for windows is an approx guess, but stopped happening
        # somewhere around 10.20220505-g3b83224e5 may be.
        # see https://github.com/datalad/datalad/issues/6849
        assert_result_count(call(), 0)
    elif ds.repo.git_annex_version < '10.20220222':
        # As of 2a8fdfc7d (Display a warning message when asked to operate on a
        # file inside a symlinked directory, 2020-05-11), git-annex will error.
        with assert_raises(CommandError):
            call()
    elif '10.20220222' <= ds.repo.git_annex_version < '10.20220322':
        # No error on annex' side since 10.20220222;
        # However, we'd now get something like this:
        # > git annex find bar/f
        # error: pathspec 'bar/f' did not match any file(s) known to git
        # Did you forget to 'git add'?
        #
        # But exists zero until 10.20220322!
        assert_result_count(call(), 0)
    else:
        res = call()
        assert_result_count(res, 1, status='error', state='unknown',
                            path=str(bar_f))


@with_tempfile
@with_tempfile
def test_get_paths_by_ds(path=None, otherdspath=None):
    otherds = Dataset(otherdspath).create()
    ds = get_deeply_nested_structure(path)

    # for testing below, a shortcut
    subds_modified = Dataset(ds.pathobj / 'subds_modified')

    # check docstrong of get_deeply_nested_structure() to understand
    # what is being tested here
    testcases = (
        # (
        #   (<dataset_arg>, <path arg>),
        #   {<path by ds dict>}
        #   [<error list>]
        # ),

        # find main dataset, pass-through arbitrary arguments, if no paths
        # go in, also no paths come out
        ((path, None), {ds.pathobj: None}, []),
        # a simple path in the rootds, stays just that, not traversal
        # into files underneaths
        ((ds, ['subdir']), {ds.pathobj: [ds.pathobj / 'subdir']}, []),
        # same for files, any number,
        # one record per dataset with multiple files
        ((ds, [op.join('subdir', 'git_file.txt'), 'directory_untracked']),
         {ds.pathobj: [ds.pathobj / 'directory_untracked',
                       ds.pathobj / 'subdir' / 'git_file.txt']},
         []),
        # same for a subdataset root -- still reported as part of
        # the superdataset!
        ((ds, ['subds_modified']),
         {ds.pathobj: [subds_modified.pathobj]},
         []),
        # but not with a trailing slash, then it is the subdataset root
        # itself that becomes the record!!!
        ((ds, ['subds_modified' + op.sep]),
         {subds_modified.pathobj: [subds_modified.pathobj]},
         []),
        # however, regardless of the path syntax, each behavior can be forced
        ((ds, ['subds_modified'], 'sub'),
         {subds_modified.pathobj: [subds_modified.pathobj]},
         []),
        ((ds, ['subds_modified' + op.sep], 'super'),
         {ds.pathobj: [subds_modified.pathobj]},
         []),
        # subdataset content is sorted into a subdataset record
        ((ds, [op.join('subds_modified', 'subdir')]),
         {subds_modified.pathobj: [ds.pathobj / 'subds_modified' / 'subdir']},
         []),
        # content from different datasets ends up in different records
        ((ds, [op.join('subdir', 'git_file.txt'),
               op.join('subds_modified', 'subdir'),
               op.join('subds_modified', 'subds_lvl1_modified')]),
         {ds.pathobj: [ds.pathobj / 'subdir' / 'git_file.txt'],
          subds_modified.pathobj: [
              subds_modified.pathobj / 'subdir',
              subds_modified.pathobj / 'subds_lvl1_modified']},
         []),
        # paths not matching existing content are no problem
        ((ds, ['doesnotexist',
               op.join('subdir', 'nothere'),
               op.join('subds_modified', 'subdir', 'gone')]),
         {ds.pathobj: [ds.pathobj / 'doesnotexist',
                       ds.pathobj / 'subdir' / 'nothere'],
          subds_modified.pathobj: [
              subds_modified.pathobj / 'subdir' / 'gone']},
         []),
        #
        # now error case
        #
        # a path that does sort under the root dataset
        ((path, [otherds.pathobj / 'totally' / 'different']),
         {},
         [otherds.pathobj / 'totally' / 'different']),
    )
    # evaluate the test cases
    for inp, pbd_target, error_target in testcases:
        paths_by_ds, errors = get_paths_by_ds(ds, *inp)
        assert_dict_equal(pbd_target, paths_by_ds)
        eq_(error_target, errors)

    # lastly, some more specialized test
    # paths get collapsed into dataset records, even when the path
    # order is not presorted to match individual datasets sequentially
    paths_by_ds, errors = get_paths_by_ds(
        ds, ds, [
            op.join('subdir', 'git_file.txt'),
            op.join('subds_modified', 'subdir'),
            op.join('subdir', 'annexed_file.txt'),
        ])
    eq_(
        list(paths_by_ds.keys()),
        [ds.pathobj, subds_modified.pathobj]
    )
    # result order (top-level first) is stable, even when a path comes first
    # that sorts later. Also mixed types are not a problem
    paths_by_ds, errors = get_paths_by_ds(
        ds, ds, [
            ds.pathobj / 'subds_modified' / 'subdir',
            op.join('subdir', 'git_file.txt'),
            op.join('subds_modified', 'subdir', 'annexed_file.txt'),
        ])
    eq_(
        list(paths_by_ds.keys()),
        [ds.pathobj, subds_modified.pathobj]
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""For now just a proxy to git annex unlock

"""

__docformat__ = 'restructuredtext'

import logging
import os.path as op
from collections import defaultdict

from datalad.core.local.status import Status
from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
    resolve_path,
)
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.interface.common_opts import (
    recursion_flag,
    recursion_limit,
)
from datalad.interface.results import get_status_dict
from datalad.log import log_progress
from datalad.support.constraints import (
    EnsureNone,
    EnsureStr,
)
from datalad.support.param import Parameter
from datalad.utils import (
    Path,
    ensure_list,
)

lgr = logging.getLogger('datalad.local.unlock')


@build_doc
class Unlock(Interface):
    """Unlock file(s) of a dataset

    Unlock files of a dataset in order to be able to edit the actual content
    """

    _params_ = dict(
        path=Parameter(
            args=("path",),
            doc="""file(s) to unlock""",
            nargs="*",
            constraints=EnsureStr() | EnsureNone()),
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc=""""specify the dataset to unlock files in. If
            no dataset is given, an attempt is made to identify the dataset
            based on the current working directory.""",
            constraints=EnsureDataset() | EnsureNone()),
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
    )

    _examples_ = [
        dict(text="Unlock a single file",
             code_py="unlock(path='path/to/file')",
             code_cmd="datalad unlock <path/to/file>"),
        dict(text="Unlock all contents in the dataset",
             code_py="unlock('.')",
             code_cmd="datalad unlock ."),
    ]

    @staticmethod
    @datasetmethod(name='unlock')
    @eval_results
    def __call__(
            path=None,
            *,
            dataset=None,
            recursive=False,
            recursion_limit=None):
        refds = require_dataset(dataset, check_installed=True,
                                purpose="unlock")

        # Before passing the results to status()
        #   * record explicitly specified non-directory paths so that we can
        #     decide whether to yield a result for reported paths
        #   * filter out and yield results for paths that don't exist
        res_paths_nondir = set()
        paths_lexist = None
        res_paths = list()
        if path:
            # Note, that we need unresolved versions of the path input to be
            # passed on to status. See gh-5456 for example.
            path = ensure_list(path)
            res_paths = resolve_path(path, ds=dataset)
            paths_lexist = []
            res_paths_lexist = []
            for p, p_r in zip(path, res_paths):
                if p_r.exists() or p_r.is_symlink():
                    paths_lexist.append(p)
                    res_paths_lexist.append(p_r)
                if not p_r.is_dir():
                    res_paths_nondir.add(p_r)

        res_kwargs = dict(action='unlock', logger=lgr, refds=refds.path)
        if res_paths:
            for p in set(res_paths).difference(set(res_paths_lexist)):
                yield get_status_dict(
                    status="impossible",
                    path=str(p),
                    type="file",
                    message="path does not exist",
                    **res_kwargs)
        if not (paths_lexist or paths_lexist is None):
            return

        # Collect information on the paths to unlock.
        to_unlock = defaultdict(list)  # ds => paths (relative to ds)
        for res in Status()(
                # ATTN: it is vital to pass the `dataset` argument as it,
                # and not a dataset instance in order to maintain the path
                # semantics between here and the status() call
                dataset=dataset,
                path=paths_lexist,
                untracked="normal" if res_paths_nondir else "no",
                annex="availability",
                recursive=recursive,
                recursion_limit=recursion_limit,
                result_renderer="disabled",
                return_type="generator",
                on_failure="ignore"):
            if res["action"] != "status" or res["status"] != "ok":
                yield res
                continue
            has_content = res.get("has_content")
            if has_content:
                parentds = res["parentds"]
                to_unlock[parentds].append(op.relpath(res["path"], parentds))
            elif res_paths_nondir and Path(res["path"]) in res_paths_nondir:
                if has_content is False:
                    msg = "no content present"
                    status = "impossible"
                elif res["state"] == "untracked":
                    msg = "untracked"
                    status = "impossible"
                else:
                    # This is either a regular git file or an unlocked annex
                    # file.
                    msg = "non-annex file"
                    status = "notneeded"
                yield get_status_dict(
                    status=status,
                    path=res["path"],
                    type="file",
                    message="{}; cannot unlock".format(msg),
                    **res_kwargs)

        # Do the actual unlocking.
        for ds_path, files in to_unlock.items():
            # register for final orderly take down
            pbar_id = f'unlock-{ds_path}'
            nfiles = len(files)
            log_progress(
                lgr.info, pbar_id,
                'Unlocking files',
                unit=' Files',
                label='Unlocking',
                total=nfiles,
                noninteractive_level=logging.INFO,
            )
            ds = Dataset(ds_path)
            for r in ds.repo._call_annex_records_items_(
                    ["unlock"],
                    files=files,
            ):
                log_progress(
                    lgr.info, pbar_id,
                    "Files to unlock %i", nfiles,
                    update=1, increment=True,
                    noninteractive_level=logging.DEBUG)
                nfiles -= 1
                yield get_status_dict(
                    path=op.join(ds.path, r['file']),
                    status='ok' if r['success'] else 'error',
                    type='file',
                    **res_kwargs)
                if nfiles < 1:
                    # git-annex will spend considerable time after the last
                    # file record to finish things up, let this be known
                    log_progress(
                        lgr.info, pbar_id,
                        "Recording unlocked state in git",
                        update=0, increment=True,
                        noninteractive_level=logging.INFO)

            log_progress(
                lgr.info, pbar_id,
                "Completed unlocking files",
                noninteractive_level=logging.INFO)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


from os.path import join as opj
from datalad.distribution.dataset import Dataset

def _make_dataset_hierarchy(path):
    origin = Dataset(path).create()
    origin_sub1 = origin.create('sub1')
    origin_sub2 = origin_sub1.create('sub2')
    with open(opj(origin_sub2.path, 'file_in_annex.txt'), "w") as f:
        f.write('content2')
    origin_sub3 = origin_sub2.create('sub3')
    with open(opj(origin_sub3.path, 'file_in_annex.txt'), "w") as f:
        f.write('content3')
    origin_sub4 = origin_sub3.create('sub4')
    origin.save(recursive=True)
    return origin, origin_sub1, origin_sub2, origin_sub3, origin_sub4


def _mk_submodule_annex(path, fname, fcontent):
    ca = dict(result_renderer='disabled')
    # a remote dataset with a subdataset underneath
    origds = Dataset(path).create(**ca)
    (origds.pathobj / fname).write_text(fcontent)
    # naming is weird, but a legacy artifact
    s1 = origds.create('subm 1', **ca)
    (s1.pathobj / fname).write_text(fcontent)
    s2 = origds.create('2', **ca)
    (s2.pathobj / fname).write_text(fcontent)
    origds.save(recursive=True, **ca)
    return origds